// (C) 2003 Ulrich Hecht <uli@emulinks.de>.
// Distributed under the terms of the GNU Public License v2.

#include <guard.h>
#include <stage.h>

/// This is the number of pixels to take every iteration. A number n greater than one means
/// that the routine takes 1/n times the time at the expense of possibly ignoring n-1 pixels
/// wide obstacles, e.g. block corners. Use with caution.
#define PIXEL_STEP 1

// uses line-drawing algorithm from http://www.gamedev.net/reference/articles/article1275.asp
bool guard::sightblocked(int spritex, int spritey, int spritexsize, int spriteysize, sprite* s, bool hasdirections)
{
	int xinc1,xinc2,yinc1,yinc2;
	int den,num,numadd,numpixels;
	int deltax,deltay;
	int x,y;
	int cornx[2];	// x coords of sprite corners
	int corny[2];	// y coords of sprite corners
	bool blocked;
	short block;
	int collmap;
	int mystartx,mystarty; // where am I looking from ("eyes position")

	bool invert;

	mystartx=getboundcenterx();
	mystarty=getboundcentery();

	if (hasdirections)
	{
		// start from "front" of guard
		ASSERT(getdirection()<=d_down, "unknown direction");
		switch (getdirection())
		{
		case d_up:
			mystarty-=getboundysize()/2;
			break;
		case d_down:
			mystarty+=getboundysize()/2;
			break;
		case d_left:
			mystartx-=getboundxsize()/2;
			break;
		case d_right:
			mystartx+=getboundxsize()/2;
			break;
		default:
			ASSERT(0, "this should never be reached");
			break;
		}
	}

	collmap=getcollidemapno();
	invert=m_parent->getmapinversion(collmap);

	cornx[0]=spritex; cornx[1]=spritex+spritexsize;
	corny[0]=spritey; corny[1]=spritey+spriteysize;

	for (int i=0;i<4;i++)	 // do the four corners
	{
		spritex=cornx[i&1];
		spritey=corny[(i&2)>>1];

		deltax = abs(mystartx - spritex);		 // The difference between the x's
		deltay = abs(mystarty - spritey);		 // The difference between the y's
		x = spritex;					   // Start x off at the first pixel
		y = spritey;					   // Start y off at the first pixel

		if (mystartx >= spritex)				 // The x-values are increasing
		{
			xinc1 = PIXEL_STEP;
			xinc2 = PIXEL_STEP;
		}
		else
		{						   // The x-values are decreasing
			xinc1 = -PIXEL_STEP;
			xinc2 = -PIXEL_STEP;
		}

		if (mystarty >= spritey)				 // The y-values are increasing
		{
			yinc1 = PIXEL_STEP;
			yinc2 = PIXEL_STEP;
		}
		else
		{						   // The y-values are decreasing
			yinc1 = -PIXEL_STEP;
			yinc2 = -PIXEL_STEP;
		}

		if (deltax >= deltay)		  // There is at least one x-value for every y-value
		{
			xinc1 = 0;					// Don't change the x when numerator >= denominator
			yinc2 = 0;					// Don't change the y for every iteration
			den = deltax;
			num = deltax / 2;
			numadd = deltay;
			numpixels = deltax/PIXEL_STEP;			// There are more x-values than y-values
		}
		else
		{						   // There is at least one y-value for every x-value
			xinc2 = 0;					// Don't change the x for every iteration
			yinc1 = 0;					// Don't change the y when numerator >= denominator
			den = deltay;
			num = deltay / 2;
			numadd = deltax;
			numpixels = deltay/PIXEL_STEP;			// There are more y-values than x-values
		}

		blocked=false;

		sprite ** sprs=m_parent->getsprites();
		int scnt=m_parent->getspritecount();
		int j;

//#define DEBUG_SIGHT		
		
		for (int curpixel = 0; curpixel <= numpixels; curpixel++)
		{
#ifdef DEBUG_SIGHT
			uprintf("checking %d/%d\n",x/8,y/8);
#endif			
			if ((block=m_parent->getmapdata(collmap,x/8,y/8))&UNOPT_BLOCK_MASK)	// block is there
			{
#ifdef DEBUG_SIGHT
				uprintf("block %d bits %d %d %d %d %d %d %d %d\n",block,blockstr[block].tl,blockstr[block].tr,blockstr[block].bl,blockstr[block].br,blockstr[block].trigger,blockstr[block].unused1,blockstr[block].unused2,blockstr[block].unused3);
#endif
				// CAVEAT EMPTOR: This is a BITWISE XOR!! Only works because both invert and half-high (aka trigger) are single bits!
				if (!invert ^ ((block&UNOPT_HALFHIGH_MASK)>>2))
				{
#ifdef DEBUG_SIGHT
					uprintf("blocked!\n");
#endif					
					blocked=true;
					break;
				}
			}

			for (j=0;j<scnt;j++)
			{
				if (sprs[j]==s || sprs[j]==this || !sprs[j]->getblockssight() || sprs[j]->getcollidemapno()!=collmap) continue;
				if (sprs[j]->collides(x,y))
				{
					blocked=true;
					break;
				}
			}
			if (j<scnt)	break;

			num += numadd;				// Increase the numerator by the top of the fraction
			if (num >= den)				// Check if numerator >= denominator
			{
				num -= den;				  // Calculate the new numerator value
				x += xinc1;				  // Change the x as appropriate
				y += yinc1;				  // Change the y as appropriate
			}
			x += xinc2;					// Change the x as appropriate
			y += yinc2;					// Change the y as appropriate
		}
		if (!blocked) return false;		// Free line of sight, I can see you!
	}
	return true;	// Sight blocked, I cannot see you.
}
