// (C) 2003 Ulrich Hecht <uli@emulinks.de>.
// Distributed under the terms of the GNU Public License v2.

#include <guard.h>
#include <stage.h>
#include <misc.h>

guard::guard(stage* parent, const spritedata* gfx, int xpos=0, int ypos=0)
: scriptable(parent, gfx, xpos, ypos)
{
	// doesn't cut it for camera objects...
	//ASSERT(animations >=4, "guard needs anims for 4 directions");

	m_fovangle=BIOSDEGREES(60);		// 120 degrees is the natural field of vision of a human being
	m_fovdepth=100;

	m_hearingdist=30;
	m_canhear=true;

	setradarcolor(S_RADAR_GUARD);
	setanimtype(sprite::a_motion);
	enablespritecollision(true);
	setblockssight(true);

	m_attractbackx=m_attractbacky=m_attractwait=0;
	m_attractwaitstart=0;
	m_attracted=0;	// attract off

	for(int i=0; i<MAX_SIGHTINGS; i++)
	{
		m_sighted[i]=0;
		m_sightedtime[i]=0; // slot not occupied
		m_sightedtoday[i]=false;
	}

	setdescription("A security guard.");
}

//#define PROFILE_GUARD
//#define DEBUG_GUARD

#ifdef PROFILE_GUARD
TOOL_PROFILER_INIT;
#endif

void guard::compute(bool lowtime)
{
	scriptable::compute(lowtime);

	if(m_attracted)
	{
		switch(m_attracted)
		{
		case 1:	// on my way
			if(!ismoving())	// we got there
			{
				m_attractwaitstart=m_parent->getticks();
				m_attracted=2;
			}
			break;
		case 2: // waiting once
			if(m_parent->getticks()>m_attractwaitstart+m_attractwait/4)
			{
				m_attractwaitstart=m_parent->getticks();
				m_attracted=3;
				setanimdirection((getdirection()-1)&3);	// turn left
			}
			break;
		case 3: // looking left
			if(m_parent->getticks()>m_attractwaitstart+m_attractwait/4)
			{
				m_attractwaitstart=m_parent->getticks();
				m_attracted=4;
				setanimdirection((getdirection()+2)&3);	// turn right twice
			}
			break;
		case 4: // looking right
			if(m_parent->getticks()>m_attractwaitstart+m_attractwait/4)
			{
				m_attractwaitstart=m_parent->getticks();
				m_attracted=5;
				setanimdirection((getdirection()-1)&3);	// turn back
			}
			break;
		case 5: // waiting twice
			if(m_parent->getticks()>m_attractwaitstart+m_attractwait/4)
			{
				m_attracted=0;
				resumescript();
				moveto(m_attractbackx, m_attractbacky, 1, 2); // FIXME: make moveback speed configurable
			}
			break;
		default:
			ASSERT(0, "should never be reached");
		}
	}
	
	if(lowtime) return;	// too much to squeeze into 50 ticks

#ifdef PROFILE_GUARD
	TOOL_PROFILER_START;
#endif
	sprite** sprites=m_parent->getsprites();
	sprite* sprite;
	int count=m_parent->getspritecount();
	int tmpx, tmpy;
	int dist;
	int angle=0;

	for (int i=0; i<count;i++)
	{
		sprite=sprites[i];
		if (sprite==this) continue;	// skip myself
		if (sprite->getcollidemapno()==getcollidemapno())
		{
			tmpx=sprite->getboundcenterx()-getboundcenterx();
			tmpy=sprite->getboundcentery()-getboundcentery();
			dist=bios_sqrt(tmpx*tmpx+tmpy*tmpy);
#ifdef DEBUG_GUARD
			uprintf("dist %d\n", dist);
#endif

			// doing hearing first coz seeing may change tmpx/tmpy values
			if (m_canhear && sprite->isaudible() && sprite->hasmovedlastturn())
			{
				if (dist<=m_hearingdist)
					if (!sightblocked(sprite->getboundxpos(), sprite->getboundypos(),sprite->getboundxsize(), sprite->getboundysize(), sprite, false))
					{
						// look into the direction the sound came from
						if (abs(tmpx)<abs(tmpy))
						{
							if (tmpy>0)	setanimdirection(d_down);
							else setanimdirection(d_up);
						}
						else
						{
							if (tmpx>0)	setanimdirection(d_right);
							else setanimdirection(d_left);
						}
						m_parent->setsighted(this,sprite);
					}
			}
			if (sprite->issightable() && dist<=m_fovdepth)
			{

				ASSERT(getdirection()<=d_down,"unknown direction");
				switch (getdirection())
				{
				case d_left:
					angle=BIOSDEGREES(180);
					tmpx+=getboundxsize()/2;
					break;
				case d_right:
					angle=BIOSDEGREES(0);
					tmpx-=getboundxsize()/2;
					break;
				case d_up:
					angle=BIOSDEGREES(-90);
					tmpy+=getboundysize()/2;
					break;
				case d_down:
					angle=BIOSDEGREES(90);
					tmpy-=getboundysize()/2;
					break;
				default:
					ASSERT(0, "this should never be reached");
					break;
				}
#ifdef DEBUG_GUARD
				uprintf("myangle %d other %d %d (a %d) diff %d fov %d\n", angle,tmpx,tmpy,bios_atan2(tmpx, tmpy),(short)(bios_atan2(tmpx, tmpy)-angle), m_fovangle);
#endif				
				if (!(abs((short)(bios_atan2(tmpx, tmpy)-angle))>m_fovangle))
					if (!sightblocked(sprite->getboundxpos(), sprite->getboundypos(),sprite->getboundxsize(), sprite->getboundysize(), sprite, true))
						registersighting(sprite);
			}
		}
	}
#ifdef PROFILE_GUARD
	TOOL_PROFILER_END;
#endif
	deregistersightings();
}

void guard::push(sprite* pusher)
{
	if (pusher->issightable())
		m_parent->setsighted(this, pusher);
	else
		sprite::push(pusher);
}

//#define DEBUG_ATTRACT

void guard::attract(int x, int y, int speed_num, int speed_denom, int waittime)
{
#ifdef DEBUG_ATTRACT
	uprintf("%s at %x attracted to %d/%d\n", getdescription(), this, x, y);
#endif
	m_attractbackx=m_xpos;
	m_attractbacky=m_ypos;
	m_attractwait=waittime;
	moveto(x, y, speed_num, speed_denom);
	suspendscript();
	m_attracted=1;	// on my way
}

void guard::registersighting(sprite* sighted)
{
	for(int i=0; i<MAX_SIGHTINGS; i++)
	{
		if(m_sightedtime[i]>0 && m_sighted[i]==sighted)
		{
			m_sightedtime[i]++;
			m_sightedtoday[i]=true;
			if(m_sightedtime[i]>20/VBL_FRAMESKIP)	// more than .333 secs -> gotcha
			{
				m_sightedtime[i]=0;
				m_parent->setsighted(this, sighted);
			}
			return;
		}
	}
	// This sprite not registered yet, try to find a new slot.
	for(int i=0; i<MAX_SIGHTINGS; i++)
	{
		if(m_sightedtime[i]==0)		// unoccupied slot
		{
			m_sightedtime[i]=1;
			m_sighted[i]=sighted;
			m_sightedtoday[i]=true;
			return;
		}
	}
	ASSERT(0, "unable to assign a sighting slot");
}

#define DEBUG_SIGHTING

void guard::deregistersightings()
{
	for(int i=0; i<MAX_SIGHTINGS; i++)
	{
		if(m_sightedtime[i]>0 && !m_sightedtoday[i])
		{
			if(m_sightedtime[i]>10/VBL_FRAMESKIP)	// more than .16667 secs -> check it out
				attract(m_sighted[i]->getboundxpos(), m_sighted[i]->getboundypos(), 1, 1, 120);
			
#ifdef DEBUG_SIGHTING
			uprintf("unsighted %s at %x at time %d\n", m_sighted[i]->getdescription(), m_sighted[i], m_sightedtime[i]);
#endif			
			m_sightedtime[i]=0;
		}
		m_sightedtoday[i]=false;
	}
}
