// (C) 2003 Ulrich Hecht <uli@emulinks.de>.
// Distributed under the terms of the GNU Public License v2.

#ifndef _DOOR_H
#define _DOOR_H

#include <sprite.h>

/// A sliding door object. Its default action is to open itself unless it is locked.
class door : public sprite
{
public:
	/// Door constructor. Identical to sprite constructor.
	door(stage* parent, const spritedata* gfx, int xpos=0, int ypos=0);

	/// Sets the position of the door in closed state.
	/// \param x X position in pixels.
	/// \param y Y position in pixels.
	void setclosedposition(int x, int y) { m_closedx=x; m_closedy=y;}

	/// Sets the position of the door in open state.
	/// \param x X position in pixels.
	/// \param y Y position in pixels.
	void setopenposition(int x, int y) { m_openx=x; m_openy=y;}

	void open();	///< Opens the door.
	void close();	///< Closes the door.
	void lock() { close(); m_locked=true;}	///< Locks the door; it will not open if someone performs an action on it.
	void unlock() { m_locked=false;}	///< Unlocks the door; it will open if someone performs an action on it.
	
	/// Set the speed the door moves at.
	/// \param speed_num Door speed numerator.
	/// \param speed_denom Door speed denominator.
	void setspeed(int speed_num, int speed_denom) { m_speed_num=speed_num; m_speed_denom=speed_denom;}

	bool isopen() { return (m_xpos==m_openx && m_ypos==m_openy)||!isvisible();}	///< \returns Is the door open?
	bool isclosed() { return (m_xpos==m_closedx && m_ypos==m_closedy && isvisible());}	///< \returns Is the door closed?

	/// A door's default action is to open itself unless locked.
	virtual void action(sprite *actor);

	/// A door's priority is lower than that of a regular sprite with the same collide map to allow it to slide under walls.
	void setcollidemapno(int map) { sprite::setcollidemapno(map); setpriority(map>0?map-1:0);}

	/// Set the amount of time after which the open door closes itself automatically. -1 disables auto-closing.
	/// \param ticks Auto-close delay in ticks.
	void setautoclose(int ticks) { m_autoclose=ticks;}

	void compute(bool lowtime);

private:
	int m_openx;		///< Open position, X value.
	int m_openy;		///< Open position, Y value.
	int m_closedx;		///< Closed position, X value.
	int m_closedy;		///< Closed position, Y value.
	int m_speed_num;	///< Movement speed, numerator.
	int m_speed_denom;	///< Movement speed, denominator.
	bool m_locked;		///< Locked?
	int m_autoclose;	///< Number of ticks to automatically close the door after. -1 means disabled.
	int m_closeat;		///< Point in time to close the door at, -1: disabled.
};

#endif
