// (C) 2003 Ulrich Hecht <uli@emulinks.de>.
// Distributed under the terms of the GNU Public License v2.

#ifndef _GUARD_H
#define _GUARD_H

#include <scriptable.h>

#define MAX_SIGHTINGS 4

/// A guardian sprite class that is scriptable and able to see sightable sprites. It also has the ability to hear the movement
/// of audible sprites. This functionality is not enabled by default. FIXME: Cannot see sprites at lower levels.
class guard : public scriptable
{
public:
	/// The guard constructor.
    guard(stage* parent, const spritedata* gfx, int xpos, int ypos);

	/// Implements the guard's abilities to see and hear in addition to the standard scriptable functionality.
	void compute(bool lowtime);

    /// Sets the angle of the field of vision.
	/// \param fovangle Angle of FOV in degrees
	void setfovangle(int fovangle) { m_fovangle=BIOSDEGREES(fovangle)/2; }
	/// Sets the depth of the field of vision.
	/// \param fovdepth Depth of FOV in pixels
    void setfovdepth(int fovdepth) { m_fovdepth=fovdepth; }

    int getfovangle() { return REALDEGREES(m_fovangle)*2; }	///< \returns Angle of FOV in degrees
    int getfovdepth() { return m_fovdepth; }				///< \returns Depth of FOV in pixels

	/// Sets the hearing distance for this guard. If hearing is enabled the guard will detect any audible moving sprite within the
	/// given hearing distance if not blocked by the map.
	/// \param dist Hearing distance in pixels.
	void sethearingdistance(int dist) { m_hearingdist=dist;}

	/// Controls the hearing ability of the guard. Set to true to enable the guard to hear moving audible sprites.
	/// \param hear Can this guard hear audible sprites moving?
	void canhear(bool hear) { m_canhear=hear;}

	/// Suspends whatever the guard may be doing and makes it moveto() the specified position, wait there for the specified time, then
	/// go back and restart the script instruction it was processing at the moment of the interruption, if any. NB: The speed parameters
	/// need not be made dependent on VBL_FRAMESKIP as this method uses a script.
	/// \param x Destination X position, in pixels
	/// \param y Destination Y position, in pixels
	/// \param speed_num Movement speed, numerator.
	/// \param speed_denom Movement speed, denominator.
	/// \param waittime Time to wait at destination, in ticks.
	void attract(int x, int y, int speed_num, int speed_denom, int waittime);

protected:
	/// Checks if line of sight to the given box (in pixels) is blocked by my collide map.
	/// \param spritex Left edge of box.
	/// \param spritey Top edge of box.
	/// \param spritesizex Horizontal size of box.
	/// \param spritesizey Vertical size of box.
	/// \param s The sprite itself. This is used to avoid a sprite to be sighted blocking sight to itself. Set to NULL if you are not checking a sprite.
	/// \param hasdirections Whether directions should be honored. If this is set to true, the line of sight starts at the "front" of the sprite
	/// (depending on direction), otherwise at the center.
    bool sightblocked(int spritex, int spritey, int spritesizex, int spritesizey, sprite* s, bool hasdirections) MEM_FUNC_IN_IWRAM;

	int m_fovangle;                     ///< Half the field of vision angle (BIOS degrees).
    int m_fovdepth;                     ///< Field of vision depth (pixels).

	int m_hearingdist;		///< Hearing distance in pixels.
	bool m_canhear;			///< Can I hear?

private:
	/// Guards sight sprites that run into them.
	virtual void push(sprite* pusher);

	/// Makes a note in the m_sighted* members that a certain sprite could be seen. Notifies the parent if a sprite could be
	/// seen for more than .333 seconds.
	/// \param sighted The sprite that can be seen.
	void registersighting(sprite* sighted);

	/// Resets entries in m_sighted* for sprites that could not be seen this turn. Calls attract() if this happens for sprites
	/// that have been seen for more than .16667 seconds.
	void deregistersightings();

	int m_attractbackx;		///< Location to move back to after reached source of attraction, X coordinate
	int m_attractbacky;		///< Location to move back to after reached source of attraction, Y coordinate
	int m_attracted;		///< Attraction state; 0 = off, 1 = on my way, 2 = waiting once, 3 = looking left, 4 = looking right, 5 = waiting twice
	int m_attractwait;		///< Number of ticks to wait at source of attraction.
	int m_attractwaitstart;	///< Point in game time at which waiting at the source of attraction started, or 0 if not waiting.

	// Sighting scratchpad
	sprite* m_sighted[MAX_SIGHTINGS];		///< Pointers to sprites that have been seen lately.
	int m_sightedtime[MAX_SIGHTINGS];		///< Amount of time the sprite could be seen continuously; slot considered unoccupied if this value is 0.
	bool m_sightedtoday[MAX_SIGHTINGS];		///< Could the sprite be seen this turn?
};

#endif

