// (C) 2003 Ulrich Hecht <uli@emulinks.de>.
// Distributed under the terms of the GNU Public License v2.

#ifndef _LEVEL_H
#define _LEVEL_H

#include <stage.h>
#include <item.h>
#include <levelplayer.h>

#define L_MAX_PLAYERS 8				///< Maximum number of level players that can be handled.
#define L_SCORE_NUMBER_XSIZE 8		///< X size of one score display number.
#define L_SCORE_NUMBER_YSIZE 16		///< Y size of one score display number.
#define L_SCORE_NUMBERS 6			///< Number of score display numbers.
#define L_SCORE_FORMAT "%6d"		///< printf() format used for score display.

#define DEBUG_LEVELS	///< Enables debug mode when select is pressed during construction

class levelplayer;
class item;

/// A game level. This differs from a stage in that it has abort conditions, an item/inventory system and other non-generic features.
/// To be able to use the inventory system you have to use levelplayer objects as your player's sprites. To allow the player to use
/// items to perform special actions you need to reimplement the playerusesitem() method in your level. Allocates a number of 16-color
/// sprites and uses 16-color object palette 15 for the score display and the inventory marker.
class level : public stage
{
	friend class levelplayer;	// levelplayer needs access to the player registration methods

public:
	/// Data type describing the current state of the level.
	enum state { c_undefined=0, c_won, c_lost, c_draw, c_abort};

	/// The level constructor.
	level(int first_bg, const mapdata* playfield);
	/// The level destructor.
	~level();

	virtual void compute(bool lowtime);

	/// Returns c_undefined when the level is still in progress, or the outcome otherwise.
	/// \returns The state of the stage.
	virtual state endofstage();

	virtual void sfxlost();		///< Produces an effect telling the player she has lost.
	virtual void sfxwon();		///< Produces an effect telling the player she has won.
	virtual void sfxdraw();		///< Produces an effect telling the player it's a draw.

	/// Sets the player who's inventory is to be used.
	void setplayer(levelplayer* player);

	levelplayer* getplayer() { return m_players[m_player];}	///< \returns The current player.

	/// Add	an item to a sprite's inventory. This request is only fulfilled if the sprite is a registered player.
	/// \param what An item.
	/// \param whos A sprite. Defaults to the current player.
	void addtoinventory(item* what, sprite* whos=0);

	/// Remove an item from a sprite's inventory. This request is only fulfilled if the sprite is a registered player.
	/// This method disables sprite collision on the dropped item.
	/// \param what An item.
	/// \param whos A sprite. Defaults to the current player.
	void removefrominventory(item* what, sprite* whos=0);

	/// Removes an item from all inventories of all registered players. Disables sprite collision on the item.
	/// \param what The item to remove.
	void removefromallinventories(item* what);

	/// Returns a pointer to the item currently selected.
	/// \returns A pointer to the item currently selected, or NULL if none is selected.
	item* getitemselected();

	/// Sets the currently selected item.
	/// \param itemno The number of an item in the current player's inventory.
	void setitemselected(int itemno);

	/// Enables players to drop items where they are standing. Disabled by default.
	/// \param drop Allow player to drop items?
	void enabledropping(bool drop) { m_droppingenabled=drop;}

	/// Called when a player performs an action on a sprite while she has an item selected. Reimplement this in your own level to enable
	/// your player(s) to use items. Make sure to return true if the item could be used successfully so it can be deselected automatically.
	/// Does nothing by default. CAVEAT: onwhat may be a NULL pointer if the player is trying to use an item
	/// on empty space.
	/// \param who The player using an item.
	/// \param what The item she is using.
	/// \param onwhat The sprite she is using it on.
	/// \returns Has the item been used successfully.
	virtual bool playerusesitem(levelplayer* who, item* what, sprite* onwhat) { return false; }

	/// Calls playerusesitem() if an item is selected, otherwise falls back to the default stage behavior, i.e. it passes the action
	/// on to the sprite acted upon.
	virtual void action(sprite* me, sprite* them);

	/// Returns the score achieved. Scoring has to be implemented in derived classes.
	/// \returns The score.
	int getscore() { return m_score;}

	/// Change the score. Adds the given value (can be negative) to the stage's score.
	/// \param delta Number to add.
	virtual void modifyscore(int delta);

	/// Returns the number of secret areas discovered. Scoring has to be implemented in derived classes.
	/// \returns The number of secrets discovered.
	int getsecrets() { return m_secrets;}

	/// Change the number of secret areas discovered. Adds the given value (can be negative) to the level's secret counter.
	/// \param delta Number to add.
	virtual void modifysecrets(int delta) { m_secrets+=delta;}

	/// Set the on-screen position of the score display.
	/// \param x Score display X position in pixels.
	/// \param y Score display Y position in pixels.
	void setscoreposition(int x, int y);

	/// Show the score display. This is the default.
	void showscore();

	/// Hide the score display.
	void hidescore();

	/// Need to ensure the palette is at most 240 colors to avoid overwriting our own palette section.
	void setspritepalette(const palettedata* palette);

	/// Check if this sprite is a registered level player.
	/// \param who A sprite.
	/// \returns Is sprite a level player?
	bool isplayer(sprite* who);

	int getclockhours() { return m_clock[3];}		///< \returns Game time passed, hours.
	int getclockminutes() { return m_clock[2];}		///< \returns Game time passed, minutes.
	int getclockseconds() { return m_clock[1];}		///< \returns Game time passed, seconds.

	/// Set the finish line. Any level players are unable to cross this line before allowcrossfinishline(true) has been called.
	/// The coordinates given must form a line orthogonal to the coordinate system.
	/// \param x1 Finish line end point one, X coordinate.
	/// \param y1 Finish line end point one, Y coordinate.
	/// \param x2 Finish line end point two, X coordinate.
	/// \param y2 Finish line end point two, Y coordinate.
	void setfinishline(u16 x1, u16 y1, u16 x2, u16 y2) {
		m_flx1=x1; m_fly1=y1; m_flx2=x2; m_fly2=y2;
	}

private:
	/// Register a player. This is automatically done for new level players. The first player registered is set as the current player.
	/// \param player A level player.
	void addplayer(levelplayer* player);

	/// Remove a player. This is automatically done when destructing level players. If a player that is set as the current player is
	/// removed, the current player will be set to none.
	/// \param player A level player.
	void removeplayer(levelplayer* player);

	/// Go to info screen mode. Allows the player to page through the descriptions of all visible objects that have one.
	void infoscreen();

	void drawinventory();	///< Redraw the current player's inventory.
	void hideinventory();	///< Hide the current player's inventory, set item selection to none.
	void setitemmarker();	///< Set (or hide) the selected item marker.

	virtual u32 spritemoved(sprite* obj, int oldx, int oldy);
	
	sprite* m_marker;		///< Selection marker sprite.
	
	bool m_gotsighted;			///< Has someone been sighted in the past?

	levelplayer* m_players[L_MAX_PLAYERS];	///< The registered players for purposes of inventory handling.
	int m_player;	///< Currently active player for purposes of inventory handling.
	int m_numplayers;	///< Number of players registered.

	int m_itemselected;		///< The number of the item currently selected, or -1 if none.
	bool m_droppingenabled;	///< Can the player drop items?

	bool m_lp;	///< "L" key pressed
	bool m_bp;	///< "B" key pressed

	int m_score;			///< Score achieved.
	int m_secrets;			///< Secrets discovered.
	sprite* m_scorenumber[6];	///< Sprites used to display the score.
	sprite* m_clocknumber[4];	///< Sprites used to display the game timer.

	u16 m_clock[4];			///< The game time clock. 0 is ticks, 1 is seconds, 2 is minutes, 3 is hours.

#ifdef DEBUG_LEVELS
	bool m_debugenabled;	///< Debug mode enabled?
#endif

	u16 m_flx1;		///< Finish line end point one, X coordinate.
	u16 m_fly1;		///< Finish line end point one, Y coordinate.
	u16 m_flx2;		///< Finish line end point two, X coordinate.
	u16 m_fly2;		///< Finish line end point two, Y coordinate.

};

#endif

