// (C) 2003 Ulrich Hecht <uli@emulinks.de>.
// Distributed under the terms of the GNU Public License v2.

#ifndef _MESSAGEBOX_H
#define _MESSAGEBOX_H

#include <widget.h>

/// Implements a message box. It adds a portrait and free text output functionality to the basic widget.
/// If you specify a portrait it uses the 16-color palette 14 in addition to the resources allocated by every widget object.
/// If modal is set to true the message box constructor will block and wait for the user to acknowledge the message
/// before deleting itself. A non-modal message box returns after creation, does not interact with the user, only writes text
/// when you call the writenextpage() method, and has to be deleted manually.
class messagebox : public widget
{
public:
	/// The messagebox constructor.
	/// \param modal Whether this is a modal dialog.
	/// \param bg The HW background to be used.
	/// \param text The text string to be displayed in the message box.
	/// \param border The border graphics.
	/// \param portrait The portrait graphics.
	/// \param xsize The horizontal size of the message box, in 8x8 tiles.
	/// \param ysize The vertical size of the message box, in 8x8 tiles.
	/// \param xpos The X position of the message box, in pixels, relative to the visible screen.
	/// \param ypos The Y position of the message box, in pixels, relative to the visible screen.
	/// \param textdelay Number of frames to pause between characters.
	messagebox(bool modal, int bg, const char* text, const spritedata* border, const portraitdata* portrait,
			   int xsize, int ysize, int xpos=0, int ypos=0, int textdelay=0);

	/// Writes the next page of text. Call this when you want your non-modal message box to display the next line
	/// of text. A modal message box handles this on its own.
	/// \returns Whether there is still text to output.
	bool writenextpage(void);

	/// Replace the text of the message box. Do not forget to call writenextpage() afterwards.
	/// \param text The new text string.
	void settext(const char* text) { clear(); m_text=text; m_position=0; m_writedone=false;}

private:
	const char* m_text;	///< The text string to be displayed.
	int m_position;	///< Current position in text.
	bool m_writedone; ///< Whether all text has been written.

	const portraitdata* m_portrait;			///< The portrait tileset.
	int m_pxsize;				///< The portrait's X size, in tiles.
	int m_pysize;				///< The portrait's Y size, in tiles.

};

#endif
