// (C) 2003 Ulrich Hecht <uli@emulinks.de>.
// Distributed under the terms of the GNU Public License v2.

#ifndef _OBJECT_H
#define _OBJECT_H

#include <mygba.h>

/// This is the base class for all on-screen objects. Positions and sizes are specified
/// in units of one pixel. This may or may not be true for derived classes.
class object
{
public:
	/// Sets the position of the object. The way the object's position affects its presentation
	/// on the screen depends on the type of object.
	virtual void setpos(int x, int y) = 0;
	void setxpos(int x) { setpos(x,m_ypos);}	///< Sets the X coordinate of the object's position.
	void setypos(int y) { setpos(m_xpos,y);}	///< Sets the Y coordinate of the object's position.
	
	/// Initiates movement of object to a certain destination at a certain speed.
	/// The actual movement is carried out by moveme() method.
	/// The object moves speed_num pixels every speed_denom frames.
	void moveto(int x, int y, int speed_num, int speed_denom);

	bool ismoving() { return m_ismoving;}	///< Returns whether the object is still moving.
	bool moveblocked() { return m_moveblocked;}	///< Returns whether the movement has been aborted due to a roadblock.
	void stop() { m_ismoving = false;}		///< Aborts movement.

	int getxpos() { return m_xpos;}		///< Returns the X coordinate of the object's position.
	int getypos() { return m_ypos;}		///< Returns the Y coordinate of the object's position.
	int getxsize() { return m_xsize;}	///< Returns the object's horizontal size.
	int getysize() { return m_ysize;}	///< Returns the object's vertical size.

	/// Perform time-intensive computations. This is intended to be used for time-consuming computations
	/// (AI etc.) the calling of which in an interrupt handler would not be a good idea. More specifically,
	/// it should contain all things that may change the game world's state. Moving a sprite is done here, but
	/// not moving a stage. compute() needs to be called manually every frame. Not calling it suspends the game's world.
	/// \param lowtime Setting this to true indicates that time is running out and that the method must complete in less than .763 ms (50 ticks on a 15.256 usec timer).
	virtual void compute(bool lowtime);

	/// Perform operations that don't consume much CPU time and need to be performed once every frame. This is called once every frame
	/// by the interrupt handler, i.e. it cannot be suspended. It should therefore not do things that may change the game's state, but only
	/// cosmetical stuff, like updating time-based animation or changing the player's viewpoint.
	virtual void vbl_handler() {}

protected:
	/// Executes movements initiated by moveto(). Should be called by compute() if the movement affects the game state (e.g. moving
	/// a sprite) or by vbl_handler() if the movement is cosmetical (e.g. changing the visible map section).
	void moveme();

	int m_xpos;			///< X position
	int m_ypos;			///< Y position
	int m_xsize;		///< X size
	int m_ysize;		///< Y size
	int m_movedeltax;	///< distance to destination (x)
	int m_movedeltay;	///< distance to destination (y)
	int m_movecountdown;	///< counting down frames until next move
	int m_movespeed_num;	///< movement speed numerator (pixels)
	int m_movespeed_denom;	///< movement speed denominator (frames)
	bool m_ismoving;		///< is object still moving?
	bool m_moveblocked;		///< Has movement been cancelled because of a roadblock? Reset by moveto().

private:	
	/// line drawing algorithm members
	int m_xinc1;		///< see http://www.gamedev.net/reference/articles/article1275.asp
	int m_xinc2;		///< see http://www.gamedev.net/reference/articles/article1275.asp
	int m_yinc1;		///< see http://www.gamedev.net/reference/articles/article1275.asp
	int m_yinc2;		///< see http://www.gamedev.net/reference/articles/article1275.asp
	int m_den;			///< see http://www.gamedev.net/reference/articles/article1275.asp
	int m_num;			///< see http://www.gamedev.net/reference/articles/article1275.asp
	int m_numadd;		///< see http://www.gamedev.net/reference/articles/article1275.asp
	int m_numpixels;	///< total number of pixels to move
	int m_curpixel;		///< current pixel

};

#endif

