// (C) 2003 Ulrich Hecht <uli@emulinks.de>.
// Distributed under the terms of the GNU Public License v2.

#ifndef _SCRIPTABLE_H
#define _SCRIPTABLE_H

#include <sprite.h>
extern "C" {
#include <lua.h>
}

#define S_MAXSCRIPTS 16

#define DEBUG_LUA	///< Enables error output and the print function

/// A LUA-scriptable sprite. This type of object can execute LUA scripts in its compute() method, providing LUA to C++ bindings for
/// the most important methods. Once a script has been started the scriptable sprite resumes the script every time compute() is called,
/// waiting for the script to yield(). Execution of the script stops either by stopping itself or if the stopscript() method is called.
/// FIXME: Document each binding.
class scriptable : public sprite
{
public:

	// Waypoint/sprite handle list tokens, shared with level editor
#include "scriptable_tokens.h"

	/// The scriptable sprite constructor.
	scriptable(stage* parent, const spritedata* gfx, int xpos, int ypos);

	/// The scriptable destructor. Deletes the object's LUA virtual machine if present.
	~scriptable();

	/// Executes the next command in the script. Apart from calling its parent class's compute() method, this implementation takes care of
	/// actually performing the scriptable's script. Derived classes can provide additional LUA -> C++ bindings using the API
	/// provided by LUA in their constructors.
    void compute(bool lowtime);

	/// Initiate processing of a given script. Every compute() call the script's execution is resumed, stopping at the next yield() call,
	/// unless one of the wait functions has been called and the condition to wait for has not been reached yet. Make sure the script
	/// has plenty of yield() calls as long script execution periods will cause the scheduler to punish your object. "Wait" functions
	/// yield implicitly.
	/// \param script An array containing a LUA script.
	/// \param size The size of the LUA script.
	/// \param data Data token array containing waypoints and sprite pointers.
	void startscript(const char* script, int size, const int* data=0);

	/// Halt the execution of the LUA script.
    void stopscript();

	/// Suspend execution of the LUA script.
	void suspendscript() { m_scriptsuspended=true;}

	/// Resume execution of the LUA script.
	void resumescript() { m_scriptsuspended=false;}

protected:
	bool m_scriptsuspended;					///< Script execution suspended?
	bool m_firstrun;						///< Is this the first run of the script?
	
private:
	// The LUA -> C++ binding methods.
	
	/// Overloaded LUA errormessage() function using uprintf for output.
	static int lua_errormessage(lua_State* L);

	/// Overloaded LUA print() function using uprintf for output.
	static int lua_print(lua_State* L);

	/// Returns the "this" pointer for the script's object. No parameters. One return value: "this" pointer.
	static int lua_this(lua_State* L);

	/// Set the position of the given sprite. Three parameters: sprite pointer, X coordinate, Y coordinate.
	static int lua_setpos(lua_State* L);

	/// Call moveto() for the given sprite. Five parameters: sprite pointer, X coordinate, Y coordinate, speed numerator,
	/// speed denominator.
	static int lua_moveto(lua_State* L);

	/// Returns whether the object is moving. One parameter: sprite pointer. One return value: is it moving?
	static int lua_ismoving(lua_State* L);

	/// Set animation loop and direction for the given sprite. Two parameters: sprite pointer, animation/direction number.
	static int lua_setanimdir(lua_State* L);

	/// Suspend the script of the given object for a certain amount of game time. Two parameters: scriptable pointer,
	/// wait time in ticks.
	static int lua_wait(lua_State* L);

	/// Suspend this script until a given object has stopped moving. One parameter: sprite pointer.
	static int lua_waitmotion(lua_State* L);

	/// Move the given sprite to a waypoint defined in the given sprite's data token array. Four parameters: scriptable pointer,
	/// waypoint number, speed numerator, speed denominator.
	static int lua_movetowp(lua_State* L);

	///	Transforms a reference to a sprite in the data token array to a pointer to that sprite. One parameter: sprite number.
	/// One return value: pointer to that sprite.
	static int lua_getsprite(lua_State* L);

	/// Show the given sprite. One parameter: sprite pointer.
	static int lua_show(lua_State* L);

	/// Hide the given sprite. One parameter: sprite pointer.
	static int lua_hide(lua_State* L);

	/// Turn the given sprite. Four parameters: sprite pointer, destination angle, speed numerator, speed denominator.
	static int lua_turnto(lua_State* L);

	/// Controls which signals cause an interrupt in the LUA script. This function causes the user-implemented LUA function
	/// sighandler() to be called with the current state of the parent stage's signals as the only parameter when one of the
	/// signals in the given mask is raised. One parameter: signal mask.
	static int lua_onsignal(lua_State* L);

	/// Raise the signals set in the given signal mask. One parameter: signal mask.
	static int lua_raisesig(lua_State* L);

	/// Suspend the given scriptable's script until one of the signals set in the given signal mask is raised. Two parameters:
	/// scriptable pointer, signal mask.
	static int lua_waitforsig(lua_State* L);

	/// Replacement for standard LUA mathlib function random() using BIOS division on the GBA.
	static int lua_random(lua_State* L);

	const char* m_script;		///< The LUA script.
	int m_scriptsize;			///< Size of the LUA script.
	const int* m_data;			///< Waypoints and sprite pointers, created by the level editor.
	int m_waitend;				///< Point in time until which scripting is suspended.
	bool m_waiting;				///< Scripting suspended until m_waitend reached?
	sprite* m_waitmotion;		///< The sprite I am waiting for to stop moving/turning, or NULL if none. In most cases this will be myself.
	lua_State* m_lua;			///< Pointer to LUA virtual machine, or NULL if no LUA script running.
	bool m_triggersignal;		///< Call LUA function onsignal() if a signal in m_triggermask is raised?
	u32 m_triggermask;			///< The signals triggering a call to LUA function onsignal()
	bool m_waitforsignal;		///< Wait for a signal in m_sigwaitingmask to be raised?
	u32 m_sigwaitingmask;		///< Signals to wait for if m_waitforsignal is set.
};

#endif

