// (C) 2003 Ulrich Hecht <uli@emulinks.de>.
// Distributed under the terms of the GNU Public License v2.

#ifndef _SPRITE_H
#define _SPRITE_H

#define MAX_ROTSETS 32	///< Maximum number of rotation sets supported (32 for GBA)

/// Converts real degrees to BIOS degrees
#define BIOSDEGREES(x) ((x)*182)

/// Converts BIOS degrees to real degrees
#define REALDEGREES(x) ((x)/182)

// Sprite priorities
#define P_ONTOP (-1)		///< Put sprite on top of all others.

extern "C" {
#include <mygba.h>
}

#include <formats.h>

#include <object.h>
#include <misc.h>

/// An A* pathfinding node.
class node
{
public:
	u16 x;		///< X position.
	u16 y;		///< Y position.
	u16 cfs;	///< Cost from start.
	u16 tc;	///< Total cost (cfs + est. cost to goal).
	struct node* parent;	///< Node where we came from.
	struct node* qparent;	///< Next node in the queue.

	/// Constructs a node for a given location
	node(u32 xx, u32 yy) MEM_FUNC_IN_IWRAM;
};

#define A_QUEUESPLIT 8	// more doesn't gain much
#define A_MAXDELQUEUES 256

/// An A* pathfinding queue. It's actually multiple queues that appear as one to the user; this trick speeds up the isin() method.
class astarqueue
{
public:
	node* queue[A_QUEUESPLIT];	///< The node queues.

	/// Contructs an empty queue.
	astarqueue()
	{
		for (int i=0; i<A_QUEUESPLIT; i++)
			queue[i]=0;
	}

	/// \returns Whether the queue is empty.
	bool isempty()
	{
		for (int i=0; i<A_QUEUESPLIT; i++)
			if (queue[i])
				return false;
		return true;
	}

	/// Deletes itself and all nodes in it.
	~astarqueue()
	{
		for (int i=0;i<A_QUEUESPLIT;i++)
			deleteme(queue[i]);
	}

	/// Start a run of queue deletion. This method deletes queues scheduled for deletion with deleteme(). It takes
	/// at most maxtime*15.256 usecs. In most cases you will want to use sprite::m_maxtime as a parameter. Run it
	/// regularly to keep A* memory usage low. Relies on GBA HW timer 2 running at 15.256 usecs.
	/// \param maxtime The maximum time to be consumed for deletion, in multiples of 15.256 microseconds.
	/// \returns False if there were no queues to be deleted, true otherwise.
	static bool deletethem(int maxtime);
	
	void nodepush(node* n) MEM_FUNC_IN_IWRAM;	///< Push a node on the queue.
	node* nodepop() MEM_FUNC_IN_IWRAM;			///< Pop the node with the least total cost.
	void removenode(u32 x, u32 y) MEM_FUNC_IN_IWRAM;	///< Remove a node from the queue. Does not delete the node!
	node* isin(u32 x, u32 y) MEM_FUNC_IN_IWRAM;	///< Checks if a node with the given coords exists in the queue.

private:
	/// Hash function for converting a pair of coordinates to a queue number. A mask/shift operation without multiplication
	/// would be marginally faster, but requires an even number of bits for the queue number.
	u32 qnum(u32 x, u32 y) { return(x*y)&(A_QUEUESPLIT-1);}
	
	/// Schedules a queue for deletion. This method takes only very small amounts of time. The actual work is done
	/// by deletethem().
	/// \param queue A* queue to be deleted.
	void deleteme(node* queue);

	static int m_delqueue;		///< Number of queues to be deleted.
	static node* m_delqueues[A_MAXDELQUEUES];	///< List of queues to be deleted.
};


/// The base class for all movable and hideable on-screen objects. Sprite objects need a parent stage to register with. Allocates one
/// HW sprite along with one VRAM graphics slot. The number and size of animations does not effect the resource allocation as animation
/// frames are copied from ROM on the fly. If A* pathing is used it will rely on GBA HW timer 2 running at 15.256 usecs. The stage class
/// takes care of handling this particular requirement.
class sprite : public object
{
	friend class stage;

public:
	/// Type of movement performed when moveto() is called. Options are either A* pathfinding or walking in a straigt line.
	enum movetype {
		t_astar, t_straight
	};

	/// Data type describing the way to animate a sprite, either time-based or motion-based.
	enum animtype {
		a_timed=0, a_motion
	};

	// shared with level editor
#include "sprite_directions.h"

	/// Data type describing the edges of the sprite's bounding box
	enum bounding_dirs {
		b_left=0, b_top, b_right, b_bottom
	};

	/// The sprite constructor.
	/// \param parent The parent stage the sprite connects to.
	/// \param gfx The sprite graphics, ordered by animations and phases.
	/// \param xpos initial X position of sprite
	/// \param ypos initial Y position of sprite
	sprite(stage* parent, const spritedata* gfx, int xpos=0, int ypos=0);

	virtual ~sprite();	///< sprite destructor.

	/// Sets the direction of the sprite. This value may be used by derived classes to implement
	/// functionality that depends on the orientation of a sprite. It is also used by the faces() method.
	void setdirection(directions direction) { m_currentdirection=direction;}

	directions getdirection() { return m_currentdirection;}	///< \return The current direction of the sprite.

	/// Sets the animation loop used for displaying this sprite.
	/// \param animation The number of the desired animation loop, starting at 0.
	void setanimation(u32 animation) { m_currentanimation=animation;}

	u32 getanimation() { return m_currentanimation;} ///< \return The current animation loop.

	/// Convenience method combining setdirection() and setanimation().
	/// \param animdir a direction or a number that corresponds to a direction
	void setanimdirection(u32 animdir)
	{
		ASSERT(animdir<=d_undefined, "unknown direction");
		setanimation(animdir); setdirection((directions)animdir);
	}

	/// Controls the way sprite animation is timed. A timed animation loop (a_timed) switches to the next animation phase
	/// each time a certain number of frames specified by the eventsperphase parameter in the constructor has elapsed. A
	/// motion-based animation loop (a_motion) changes animation phases every time the sprite has moved a certain number
	/// of pixels specified by the eventsperphase parameter.
	/// \param a Type of animation.
	void setanimtype(animtype a) { m_animtype=a;}

	/// Start the animation loop from the beginning.
	void resetanimation() { m_currentphase=0; m_eventcount=m_eventsperphase;}

	void setpos(int x, int y);

	void setxpos(int x) { setpos(x,getypos());}	///< Sets the X coordinate of the object's position.
	void setypos(int y) { setpos(getxpos(),y);}	///< Sets the Y coordinate of the object's position.

	int getboundxpos() { return m_xpos+m_bounding[b_left];}	///< \return The left edge of the bounding box relative to the stage origin.
	int getboundypos() { return m_ypos+m_bounding[b_top];}		///< \return The top edge of the bounding box relative to the stage origin.
	int getboundxsize() { return m_xsize-m_bounding[b_left]-m_bounding[b_right];}	///< \return The width of the bounding box.
	int getboundysize() { return m_ysize-m_bounding[b_top]-m_bounding[b_bottom];}	///< \return The height of the bounding box.
	int getboundcenterx() { return getboundxpos()+getboundxsize()/2;}	///< \return The bound center of the sprite, X coordinate.
	int getboundcentery() { return getboundypos()+getboundysize()/2;}	///< \return The bound center of the sprite, Y coordinate.

	/// Controls whether this sprite collides with its collide map. Note that this setting does not affect A* pathing.
	/// \param collide Map collision on/off.
	void enablemapcollision(bool collide) { m_collide=collide;}

	bool getmapcollision() { return m_collide;}	///< \returns State of map collision for this sprite.

	/// Sets the map a collidable sprite will collide with. Defaults to 0. This map is also used by A* pathfinding and may be used
	/// by derived classes for various purposes. Further, only sprites with the same collide map will collide with each other.
	/// \param map The map number to be used for sprite collision.
	void setcollidemapno(int map);

	/// Returns the collision map used.
	/// \returns The number of the map used for sprite collision.
	int getcollidemapno() { return m_collidemap;}

	/// Sets the sprite's priority. Normally a sprite has the same priority as its collide map, i.e. it sits on top of that map.
	/// This method allows you to change that.
	/// \param prio Map on top of which the sprite is supposed to reside.
	void setpriority(int prio);

	/// Call this method to make sure that this sprite is displayed on top of the other sprite if they have the same priority.
	/// \param other The other sprite.
	void setobjbefore(sprite* other);
	
	/// Controls sprite to sprite collision. The engine will not allow two or more sprites with sprite collision enabled to occupy the
	/// same space. If a sprite-collidable sprite tries to move "over" another the push() method of the latter will be called
	/// which can be reimplemented in derived classes so appropriate action can be taken.
	/// \param collide Sprite collision on/off.
	void enablespritecollision(bool collide) { m_spritecollide=collide;}

	bool spritecollisionenabled() { return m_spritecollide;}	///< \returns State of sprite collision for this sprite.

	/// Checks whether this sprite collides with a given map location. Always returns false if the sprite has sprite collision
	/// disabled.
	/// \param x X coordinate.
	/// \param y Y coordinate.
	/// \returns Whether the sprite occupies the position.
	bool collides(int x, int y)
	{
		if (!m_spritecollide) return false;

		int mx1=getboundxpos();
		int mx2=mx1+getboundxsize();
		int my1=getboundypos();
		int my2=my1+getboundysize();

		if ( (x>=mx1 && x<=mx2) && (y>=my1 && y<=my2) )
			return true;
		else
			return false;
	}

	/// Controls whether this sprite blocks other sprite's sight. Derived classes able to "see" use this flag to decide whether they can
	/// see through this sprite or not.
	/// \param blocks Block sight?
	void setblockssight(bool blocks) { m_blockssight=blocks;}

	/// \returns Whether this sprite blocks other sprite's sight.
	bool getblockssight() { return m_blockssight;}

	/// Sets the width for the autoadjust feature.
	/// Autoadjust enables enjin to automatically adjust the parameters passed to setpos() so that no
	/// foreground collision occurs. This is to avoid the necessity for the sprite to be perfectly aligned when it
	/// e.g. tries to get through a tight passageway.
	/// \param adjust The autoadjust width in pixels.
	void setadjust(u32 adjust) { m_adjust=adjust;};

	/// Sets the speed for the autoadjust feature.
	/// This value indicates how many pixels the sprite's position is adjusted by each time it hits an obstacle.
	/// Defaults to one.
	/// \param adjustspeed The autoadjust speed in pixels.
	void setadjustspeed(u32 adjustspeed) { m_adjustspeed=adjustspeed;}

	u32 getadjust() { return m_adjust;}	///< \returns The current autoadjust width in pixels.
	u32 getadjustspeed() { return m_adjustspeed;}	///< \returns The current autoadjust speed in pixels.

	/// Sets the dimensions of the sprite's bounding box. The bounding box is used for
	/// collision detection. Some derived classes use it for other purposes as well.
	/// \param bd The bounding box edge to be modified.
	/// \param pixels The width in pixels of the space between the corresponding edge of the sprite and the given edge of the bounding box
	void setbounding(bounding_dirs bd,int pixels) { m_bounding[bd]=pixels;};

	/// Sets whether this sprite can be seen by other sprites.
	/// This flag is used by instances of derived classes to decide whether to ignore this sprite or not.
	/// \param sightable Visibility of this sprite to other sprites.
	void setsightable(bool sightable) { m_sightable=sightable;}

	/// Sets whether this sprite can be heard by other sprites when moving.
	/// This flag is used by instances of derived classes to decide whether to ignore motion of this sprite or not.
	/// \param audible Audibility of this sprite.
	void setaudible(bool audible) { m_audible=audible;}

	bool issightable() { return m_sightable;}	///< \returns Whether you should care about this sprite.
	bool isaudible() { return m_audible; }		///< \returns Whether you can hear this sprite.
	bool hasmovedlastturn() { return m_hasmovedlastturn;}	///< \returns Whether this sprite has changed its position last turn.

	/// Returns the sprite I am looking at. Returns a pointer to the sprite this sprite is looking at, or NULL if no sprite registered with
	/// the parent stage is in front of this one, or if all are farther away than the face tolerance value permits. Uses the direction
	/// of the sprite.
	/// \return A pointer to a sprite, or NULL.
	sprite* faces();

	/// Sets the face tolerance in pixels used by the faces() method.
	/// \param facetolerance The face tolerance in pixels.
	void setfacetolerance(int facetolerance) { m_facetolerance=facetolerance;};

	int getfacetolerance() { return m_facetolerance;}	///< \return The current face tolerance.

	void show() { m_visible=true; ham_SetObjVisible(m_objno,1);}	///< Turns on on-screen display of this sprite.
	void hide() { m_visible=false; ham_SetObjVisible(m_objno,0);}	///< Turns off on-screen display of this sprite.
	bool isvisible() { return m_visible;}	///< \returns Is this sprite currently visible?

	/// Reposition me to current position. Needs to be called if something happened outside this object that requires this sprite
	/// to be repositioned, e.g. if the parent stage has changed its position
	void reposition();

	/// Sets the color representing this sprite on the radar screen. Setting it to zero makes the sprite invisible on the radar.
	/// \param col The desired color, or zero.
	void setradarcolor(int col) { m_radarcolor=col;}

	/// Returns the color this sprite has on the radar screen.
	/// \returns Radar color.
	int getradarcolor() { return m_radarcolor;}

	/// Send sprite on a journey. This method needs few CPU cycles even if you have A* pathing enabled; all the hard work is done
	/// by compute(). The only exception is if you have chosen A* movement and call it right after the pathfinding has finished.
	/// In that case it will have to delete the old queues first. Using A* pathing requires the sprite to have at least four
	/// different animation loops to represent the four basic directions.
	void moveto(int x, int y, int speed_num, int speed_denom);

	/// A sprite's compute() method must honor the lowtime parameter, i.e. if lowtime is true it must not take longer than .763 ms (50
	/// ticks on a 15.256 usec timer). Keep this in mind when adding functionality to this method in derived classes. If you compile with
	/// DEBUG_LOWTIME a warning will be shown if a compute() method exceeds this time limit.
	virtual void compute(bool lowtime);

	bool ismoving() { return(m_ismoving || m_ispathing || m_resumepathing);}
	void stop();

	/// Controls the type of movement performed when moveto() is called. Defaults to walking in a straight line.
	/// \param m A type of movement.
	void setmovetype(movetype m)
	{
		ASSERT(m!=t_astar || m_animations >=4, "A* needs at least 4 anim loops");
		m_movetype=m;
	}

	/// \returns The type of movement performed when moveto() is called.
	movetype getmovetype() { return m_movetype;}

	/// Controls how many nodes the A* pathfinding is allowed to construct at most. One node is 16 bytes. Defaults to 1024, i.e.
	/// 16 kB.
	/// \param count Number of nodes.
	void setastarmaxnodecount(int count) { m_maxnodecount=count;}

	/// \returns Number of nodes that may be used by A* pathing.
	int getastarmaxnodecount() { return m_maxnodecount;}

	/// Controls the time in percent of one frame (i.e. 1/6000 seconds) after which the A* pathing yields. Use this to make the time
	/// consumed by the compute() method during path planning more manageable. Defaults to 35 times VBL_FRAMESKIP. See stage::compute()
	/// for an explanation of the significance of this setting.
	/// \param time Time in percent of one frame (1/6000 seconds)
	void setastarmaxtime(int time) { m_maxtime=time*11; m_maxtimepercent=time; /* Assumes we're using a 15.256 usec timer */ }

	/// \returns Maximum A* pathing time in percent of one frame (1/6000 seconds) per compute() call.
	int getastarmaxtime() { return m_maxtimepercent;}

	/// Controls whether a sprite can be rotated. On the GBA enabling rotation allocates one of 32 object rotation sets.
	/// \param rot Rotation enabled?
	void enablerotation(bool rot);

	bool isrotatable() { return m_rotatable;}	///< \returns Whether this sprite is rotatable.

	/// Changes the rotation angle of this sprite to the given value instantaneously.
	/// \param angle Rotation angle in degrees.
	void setangle(int angle);

	int getangle() { return m_angledeg;}	///< \returns Current rotation angle in degrees.

	/// Initiates a turn to a given angle.
	/// \param angle Destination angle.
	/// \param speednum Number of degrees to move per step.
	/// \param speeddenom Number of compute() calls one step should take.
	void turnto(int angle, int speednum, int speeddenom);

	/// Stops the turn currently in progress.
	void stopturn() { m_turning=false;}

	bool isturning() { return m_turning; }	///< \returns Whether this sprite is currently rotating.

	/// Called to try to perform an implementation-dependant action. Use this e.g. when the player stands in front of this sprite and presses
	/// the "action" button. The default behavior is to do nothing.
	/// \param actor The sprite that tries to perform the action.
	virtual void action(sprite* actor) {}

	const char* getdescription() { return m_description;}	///< \returns The description string for this object.

	/// Sets the sprite's description string.
	void setdescription(const char* descr) { m_description=descr;}

	/// Makes the sprite's position relative to the visible screen if screenrelative is true, and relative to the map
	/// origin if screenrelative is false.
	/// \param screenrelative Frame of reference is screen?
	void setframeofreferencetoscreen(bool screenrelative) { m_screenrelative=screenrelative;}

	bool isscreenframeofreference() { return m_screenrelative;} ///< \returns Whether the visible screen is the frame of reference.

	/// Checks whether this sprite collides with another sprite.
	/// \param other The sprite to check.
	/// \returns Collision or not.
	bool collides(sprite* other)
	{
		// both me and the other sprite need to be visible, have sprite collision enabled and need to be on the same map layer
		if (!m_spritecollide || !other->isvisible() || !other->spritecollisionenabled() || m_collidemap != other->getcollidemapno()) return false;

		int ox1=other->getboundxpos();
		int ox2=ox1+other->getboundxsize()-1;
		int oy1=other->getboundypos();
		int oy2=oy1+other->getboundysize()-1;

		int mx1=getboundxpos();
		int mx2=mx1+getboundxsize()-1;
		int my1=getboundypos();
		int my2=my1+getboundysize()-1;

		if(overlap(ox1,oy1,ox2,oy2,mx1,my1,mx2,my2))
			return true;
		else
			return false;
	}

	/// Initializes the sprite palette with the given data. This method is only available on GBA.
	/// \param palette Palette data.
	static void setpalette(const palettedata* palette);

	/// Initializes a 16-color sprite palette with the given data. This method is only available on GBA.
	/// \param palno Number of palette (0-15).
	/// \param palette Palette data.
	static void setpalette16(int palno, const palettedata* palette);

	/// Make the object react to other objects moving into it, thus making it possible to push it around.
	/// \param shoveit Should the object react to pushes by moving?
	void setshovable(bool shoveit) { m_shovable=shoveit;}

	bool getshovable() { return m_shovable;}	///< \returns Whether the object can be pushed about.

protected:
	/// Called to inform the sprite that another sprite has pushed it, i.e. ran into it. The default behaviour is to ignore this.
	/// \param pusher The sprite that pushed me.
	virtual void push(sprite* pusher);

	stage* m_parent;	///< The parent stage.
	int m_objno;		///< HAM object number
	u16 m_angle;		///< Rotation angle in BIOS degrees. This is faster than using real degrees because it has implicit modulo arithmetics.
	u16 m_angledeg;		///< Rotation angle in real degrees. This is kept to avoid having to convert back and forth as this might need division.
	bool m_visible;		///< Sprite currently visible?

private:
	/// Allocate an object rotation set. This method is only available on GBA.
	/// \returns An object rotation set.
	static int getrotset();

	/// Free an object rotation set allocated with getrotset(). This method is only available on GBA.
	/// \param rotset Number of the rotation set as returned by getrotset().
	static void freerotset(int rotset);

	/// Take the next step in the A* path.
	void nextstepinpath();

	// A* pathfinding stuff, see sprite_astar.arm.cpp for docs.
	u16* findpath(u32 x, u32 y, u32 dx, u32 dy) MEM_FUNC_IN_IWRAM;
	bool stepisvalid(u32 x, u32 y, u32 dx, u32 dy) MEM_FUNC_IN_IWRAM;

	u16* m_path;			///< The A* path. Contains a number of coordinate pairs describing the path to the destination, terminated with 0xffff/0xffff.
	bool m_ispathing;		///< Whether I am still following the A* path.
	int m_pathstep;			///< The current section of the A* path.
	int m_nodecounter;		///< Counts the nodes used for A* pathing.
	int m_maxnodecount;		///< Max. number of nodes we are allowed to use.
	int m_maxtime;			///< Max. ticks of one frame to use for A* per compute() call
	int m_maxtimepercent;	///< Max. percent of one frame to use for A* per compute() call
	bool m_resumepathing;	///< Should I resume pathing on the next compute() call?
	astarqueue* m_open;		///< A* "open" queue.
	astarqueue* m_closed;	///< A* "closed" queue.
	bool m_deleteopen;		///< Delete "open" queue on next compute() call?
	bool m_deleteclosed;	///< Delete "closed" queue on next compute() call?
	u32 m_pathxdest;		///< A* destination (X)
	u32 m_pathydest;		///< A* destination (Y)
	bool m_honorsprites;	///< Check for blocking objects when pathfinding. Set by compute(), reset by moveto() after pathing has finished.
	
	/// If nextstepinpath() encounters an obstacle it requests a repath by setting m_repath to true. The next compute() call will then
	/// start another pathfinding run that also checks objects (see m_honorsprites) and reset m_repath.
	bool m_repath;

	/// Here the position of the sprite before the last nextstepinpath() run is saved. This position is used to backtrack if an unexpected
	/// obstacle is hit while walking the path.
	int m_prevx,m_prevy;
	
	movetype m_movetype;	///< Movement type for moveto().

	/// Does things that have to be done during VBlank. Should be called by the parent. Should only do things that do not change the
	/// game world, e.g. moving the stage.
	void vbl_handler();

	void redraw();	///< Updates the animation, called by vbl_handler()

	const u8* m_graphics;	///< pointer to the raw graphics data
	u32 m_animations;	///< number of animation loops
	u32 m_phases;		///< number of animation phases per loop
	u32 m_currentphase;	///< current animation phase
	directions m_currentdirection;	///< current sprite direction
	u32 m_currentanimation;	///< current animation loop
	u32 m_eventsperphase;	///< frames per animation phase
	int m_eventcount;		///< countdown to next animation phase
	u32 m_lastanimation,	///< animation loop number after last redraw() call
	m_lastphase;		///< animation phase number after last redraw() call
	animtype m_animtype;	///< Type of animation, timed or motion-based.
	u32 m_colormode;		///< 1 is 256 colors, 0 is 16 colors

	bool m_collide;		///< Map collision enabled?
	int m_collidemap;	///< Number of map used for map collision detection.
	bool m_spritecollide; ///< sprite collision enabled?

	bool m_blockssight;	///< Am I blocking other sprite's sight?

	bool m_sightable;	///< Sprite sightable by other sprites?
	bool m_audible;		///< Can sprite be heard by other sprites when moving?
	bool m_hasmoved;	///< Has the sprite moved this round?
	bool m_hasmovedlastturn; ///< Has the sprite moved during the preceding compute() round?

	u32 m_adjust;		///< autoadjust value
	u32 m_adjustspeed;	///< autoadjust speed in pixels
	int m_bounding[4];	///< bounding box edges
	int m_facetolerance;	///< tolerance within which a sprite is considered by the faces() method

	int m_radarcolor;	///< The color of the sprite on the radar screen.

	static bool m_rotsets[MAX_ROTSETS];	///< Array containing allocation info for object rotsets.

	bool m_rotatable;	///< Whether rotation is enabled for this object.
	int m_rotset;		///< Rotset allocated for this object.

	bool m_turning;			///< turnto() turning in progress?
	int m_turntoangle;		///< BIOS degrees to go before hitting destination angle.
	int m_turnspeednum;		///< Turn speed nominator in BIOS degrees.
	int m_turnspeednumdeg;	///< Turn speed nominator in real degrees.
	int m_turnspeeddenom;	///< Turn speed denumerator.
	int m_turncountdown;	///< Countdown to next turn step.

	const char* m_description;	///< The sprite's description.

	bool m_screenrelative;	///< Sprite's position relative to the screen origin (as opposed to the map origin)?

	bool m_shovable;		///< Can the sprite be pushed around?
};

#endif

