// (C) 2003 Ulrich Hecht <uli@emulinks.de>.
// Distributed under the terms of the GNU Public License v2.

#ifndef _STAGE_H
#define _STAGE_H

extern "C" {
#include <mygba.h>
#include <gbamappy.h>
}

#include <formats.h>

#include <object.h>
#include <sprite.h>
#include <vbl.h>

// if you plan to change this, be prepared to rewrite the map fragment inserter(s)
// in stage::setpos(), or use the dog-slow ham_InsertMapFragment
#define MAP_SIZE 64         ///< total screen map size (64 = 512*512)
#define MAP_MODE 3          ///< 3=64x64, 0=32x32, must fit MAP_SIZE
#define MAP_SIZE_HALF (MAP_SIZE/2)  ///< size of one map section
#define MAP_SIZE_P (MAP_SIZE*8)     ///< map size in pixels
#define MAP_SIZE_HP (MAP_SIZE_HALF*8)   ///< size of one map section in pixels

#define MAX_OBJECTS 128         ///< maximum number of sprites belonging to one stage instance

#define SCREEN_WIDTH 240		///< width of the hardware screen
#define SCREEN_HEIGHT 160		///< height of the hardware screen

#define MAPDATA m_mapdata_unopt
#define XYTOMAP(x, y, sx) ((x)+(y)*(sx))

// radar definitions
extern const spritedata circle_sprite;
#define S_RADAR_OFF (0)
#define S_RADAR_PLAYER (3)
#define S_RADAR_GUARD (4)
#define S_RADAR_CAM (5)
#define S_RADAR_SWAG (6)
#define S_RADAR_ITEM (7)

#define S_MAXHIGHLIGHTS 4	///< Maximum number of highlights that can be set at a time.

#define S_MAXCAMHINTS 8		///< Maximum number of camera hints.

/// The camera hint structure.
typedef struct
{
	/// Edges of the trigger area.
	u16 l,t,b,r;
	/// Hinted position of the camera.
	u16 camx,camy;
} camhint;

/// A generic stage, consisting of a map with one or more layers representing the scenario and a number of sprite objects. The stage class
/// is supposed to represent the current stage or level or scenario in its entirety and serves as a container for all sprites appearing on it.
/// It also provides a radar screen and is able to highlight certain map sections or objects on demand. Uses GBA HW timer 2 for scheduling.
class stage : public object
{
	friend class sprite;

public:
	/// The stage constructor.
	/// \param first_bg The first hardware background to be used for this stage.
	/// \param playfield The map data. The nature of this data is implementation-dependant.
	stage(int first_bg, const mapdata* playfield);

	virtual ~stage();	///< The stage destructor.

	void setpos(int x, int y);

	int getnumlayers() { return m_num_bgs;}	///< \returns The number of map layers used.

	/// Controls how much of the right and bottom borders of the maps is considered invalid data. Maps have to have certain dimensions
	/// (multiples of 32) to allow for efficient rendering. If you had to pad your maps to meet these constraints, use this method to
	/// tell how much of the map data should be hidden from the player.
	/// \param l Excess in tiles at the left border.
	/// \param t Excess in tiles at the top border.
	/// \param r Excess in tiles at the right border.
	/// \param b Excess in tiles at the bottom border.
	void setexcess(int l, int t, int r, int b)
	{
		m_lexcess=l; m_texcess=t;
		m_rexcess=r; m_bexcess=b;
		setpos(m_xpos, m_ypos);	 // make sure we're at a valid position
	}

	/// Returns the amount of map data cut at the left edge.
	/// \returns Excess at left edge.
	int getlexcess() { return m_lexcess;}

	/// Returns the amount of map data cut at the top edge.
	/// \returns Excess at top edge.
	int gettexcess() { return m_texcess;}

	/// Returns the amount of map data cut at the right edge.
	/// \returns Excess at right edge.
	int getrexcess() { return m_rexcess;}

	/// Returns the amount of map data cut at the bottom edge.
	/// \returns Excess at bottom edge.
	int getbexcess() { return m_bexcess;}

	/// Locks the stage to a certain sprite. The stage tries to move itself in such a way that the locked-on sprite is always shown
	/// within the given scroll border.
	/// \param obj The object to lock the stage on.
	void scrolllock(sprite* obj);
	void scrollunlock() { m_scrolllocked=-1;}	///< Unlocks the stage scrolling.
	void setscrollborderx(int scrollborder) { m_scrollborderx=scrollborder;}	///< Sets the width of the scroll border.
	void setscrollbordery(int scrollborder) { m_scrollbordery=scrollborder;}	///< Sets the height of the scroll border.

	/// Returns a pointer to the stage's sprite list. FIXME: This should only be accessible to children of sprite.
	/// \returns My sprite list.
	sprite** getsprites() { return m_sprites;}
	/// Returns the number of sprites registered with this stage. FIXME: This should only be accessible to children of sprite.
	/// \return My sprite count.
	int getspritecount() { return m_spritecnt;}

	/// Loads the palette to be used for all sprite objects. This method is only available on GBA.
	/// \param palette The sprite palette.
	void setspritepalette(const palettedata* palette);

	/// Sets the maximum amount of time the compute() method is allowed to take, in percent of one frame (1/6000 seconds). Defaults to 90
	/// times VBL_FRAMESKIP.
	void setmaxcomputetime(int time) { m_maxcomputetime=(time-10*VBL_FRAMESKIP)*11;}

	/// When this method is called all sprites registered up to that point will have a fixed priority, i.e. they will never be relegated
	/// to a lower priority. By default the first sprite has fixed priority.
	void setfixedpriority() { m_fixedprio=m_spritecnt;}

	/// Calls the compute() methods of its sprites, draws the radar and has unused A* pathing queues that may still be around deleted.
	/// It tries to schedule these calls so that the maximum number of time as set with setmaxcomputetime() is not exceeded. In doing
	/// this it relies on the assumption that no single compute() of any of its children takes longer than the allotted time. In the
	/// case that a sprite's compute() method's execution time suddenly rises by a large amount during a high-load situation the execution
	/// time of this method may still exceed the allotted time slightly. See setprofiletime() for a way to lessen this effect. Also, derived
	/// classes' compute() methods may block in order to suspend the game's world. A stage's compute() method need not honor the lowtime
	/// parameter.
	void compute(bool lowtime);
	
	/// Returns the game time counter. This is an internal counter incremented each time compute() is called. It allows you to stage
	/// timed events, e.g. animation. Remember this counts game time as opposed to real time. For the latter, refer to vbl::getframecount().
	/// This counter does not depend on VBL_FRAMESKIP.
	/// \returns Current game time counter value.
	int getticks() { return m_ticks;}	

	/// Inform the stage that a sprite has been sighted by another sprite. FIXME: This should only be accessible to children of sprite.
	/// \param me The sprite that has sighted.
	/// \param him The sprite that was sighted.
	void setsighted(sprite* me, sprite* him) { m_sighted=him; m_sightedby=me;}

	/// Returns a pointer to the sprite that has been sighted.
	/// \returns Pointer to the sighted sprite, or NULL if none.
	sprite* getsighted() { sprite* tmp=m_sighted; m_sighted=NULL; return tmp;} // FIXME: this is racy
	/// Returns a pointer to the sprite that has sighted a sightable sprite.
	/// \returns Pointer to the sighting sprite.
	sprite* getsightedby() { return m_sightedby;}
	
	/// Replace the current map with a different one. You can use this to reveal hidden passages and the like.
	/// \param newmap The new map to be used. Needs to have the same size and number of layers as the current one.
	void replacemapdata(const mapdata* newmap);
	
	/// Get the block value of a certain tile in a map.
	/// \param bg The number of the map.
	/// \param x The X coordinate of the tile
	/// \param y The Y coordinate of the tile
	/// \returns The value of the tile.
	short getmapdata(int bg, int x, int y) { return MAPDATA[bg][XYTOMAP(x, y, m_xsize)];}

	/// Get a pointer to the BLKSTR of a certain block
	/// \param tile The block number.
	/// \returns A pointer to the BLKSTR of this block.
	BLKSTR* getmapblockstr(int tile) { return &m_blockstr[tile];}

	/// Set the collision detection inversion for a given map in this stage. Normally, collidable sprites collide with non-zero
	/// map tiles. When inversion is enabled, they collide with zero (transparent) map tiles.
	/// \param map One of this stage's maps.
	/// \param invert Inversion on/off.
	void setmapinversion(int map, bool invert);
	/// Returns the inversion state of a given map.
	/// \param map One of this stage's maps.
	/// \returns Inversion status of the given map.
	bool getmapinversion(int map) { return m_mapinvert[map];}

	/// Raise one or more signals. This method raises one or more of a set of stage-global signals. Each bit in the mask
	/// corresponds to one signal, i.e. there is a total of 32 signals.
	/// \param signal The signal mask.
	void raisesignal(u32 signal) { m_signalsnext |= signal;}
	/// Drop one or more signals. This method drops one or more signals.
	/// \param signal The signal mask.
	void dropsignal(u32 signal) { m_signalsnext &= ~signal;}
	/// Returns the state of one or more signals. If multiple bits are set in the signal mask, the return value will be true if
	/// at least one signal is set.
	/// \param signal The signal mask.
	/// \returns The state of the signal(s).
	bool israised(u32 signal) { return(bool)(m_signals&signal);}

	/// Get the current state of all stage signals.
	/// \returns The signal map.
	u32 getallsignals() { return m_signals;}

	/// Sets up the radar screen. This feature allocates one background, a number of BG tiles, and 16 color BG palette no. 15.
	/// It also uses the special effects.
	/// The resources are not claimed until after you call enableradar(true), and are freed when calling enableradar(false), or
	/// on destruction of the object. Note that this feature needs large amounts of CPU time (about 30 scanlines per frame).
	/// \param bg The HW background to be used.
	/// \param x The on-screen X position.
	/// \param y The on-screen Y position.
	/// \param xsize The horizontal size in pixels (must be multiple of 8).
	/// \param ysize The vertical size in pixels (must be multiple of 8).
	/// \param pal A 16-color palette, defaults to a standard palette.
	void setradar(int bg, int x, int y, int xsize, int ysize, const palettedata* pal=&m_radardefpal);

	/// Enables or disables the radar screen. The radar has to be initialized with setradar() before you can enable it. Enabling
	/// or disabling the radar allocates or frees the resources used by it.
	/// \param enable Enable the radar?
	void enableradar(bool enable);

	bool radarenabled() { return m_radarenable;}	///< \returns Whether the radar screen is enabled.

	/// Centers on a given point on the map and highlights it using the given 32x32 16-color bitmap as a mask. It allocates a 32x32
	/// sprite and uses the object window and special effects.
	/// \param x X coordinate of map position to highlight, in pixels
	/// \param y Y coordinate of map position to highlight, in pixels
	/// \param bitmap A 16-color 32x32 tileset, defaults to a circle.
	/// \param centerx X coordinate of the on-screen position of the center of the highlight.
	/// \param centery Y coordinate of the on-screen position of the center of the highlight.
	void addhighlight(int x, int y, int centerx=SCREEN_WIDTH/2, int centery=SCREEN_HEIGHT/2, const spritedata* bitmap=&circle_sprite);

	/// A convenience method that can be used to highlight a certain object.
	/// \param s A sprite on the current stage.
	/// \param centerx X coordinate of the on-screen position of the center of the highlight.
	/// \param centery Y coordinate of the on-screen position of the center of the highlight.
	void addhighlightsprite(sprite* s, int centerx=SCREEN_WIDTH/2, int centery=SCREEN_HEIGHT/2);

	/// Disables all highlights and frees the resources used.
	void hidehighlights();

	/// Called by any sprite that tries to perform an action on another sprite. This method in turn calls the action()
	/// method of the second sprite. Derived classes may implement functionality beyond that.
	/// \param me The sprite trying to perform an action.
	/// \param them The sprite the action should be performed on.
	virtual void action(sprite* me, sprite* them);

	/// Sets the default collide map assigned to new sprites
	/// \param colmap Default collide map number.
	void setdefaultcollidemap(int colmap);

	/// Retrieves an element of the static sprite list. This is used by scriptable sprites when calling getsprite() in a LUA
	/// script.
	/// \param num The sprite number.
	/// \returns The sprite registered under that number.
	sprite* getstaticsprite(int num) { return m_staticsprites[num];}

	/// Add a camera hint. If the scrolllocked sprite walks into the map rectangle specified by l/t/r/b the camera is locked
	/// at the position given by camx/camy.
	/// \param l Left edge of camera hint rectangle.
	/// \param t Top edge of camera hint rectangle.
	/// \param r Right edge of camera hint rectangle.
	/// \param b Bottom edge of camera hint rectangle.
	/// \param camx Hinted position, X coordinate.
	/// \param camy Hinted position, Y coordinate.
	void addcamhint(u16 l, u16 t, u16 r, u16 b, u16 camx, u16 camy);

protected:
	void vbl_handler();

	/// Forces the given coordinates to a valid stage position. setpos() uses this function for all coordinates given to it,
	/// so you only need to call this manually if you need to know the final coordinates beforehand. See spritemoved() for an example.
	/// \param x pointer to an X coordinate
	/// \param y pointer to a Y coordinate
	void fixpos(int* x, int* y);

	/// Set the profiling time for a given sprite. If you know that a sprite will have a sudden rise in profile at one point you
	/// can use this method to warn the scheduler about it in advance.
	/// \param s A sprite registered with this stage.
	/// \param time The profiling time in percent of one frame (1/6000 seconds).
	void setprofiletime(sprite* s, int time);

	int getspriteprofile(int num) { return m_spriteprofile[num];} ///< \returns The profile of the sprite with the given list position.

	/// Deletes all registered sprites. Usually called from the stage dtor, this method must be used in dtors of derived classes if
	/// if there are sprites that use their interface.
	void deleteallsprites();

	/// Sets an element of the static sprite list. This list is used by scriptable sprites to make reference to fellow sprites.
	/// In order for this to work every sprite referenced has to register itself with the static sprite list with a unique number.
	/// Levels created with the level editor take care of that.
	/// \param s The sprite to register.
	/// \param num Number to register it as.
	void setstaticsprite(sprite* s, int num) { m_staticsprites[num]=s;}

	/// Informs the stage that a sprite has moved. If the sprite is collidable the return value will be either 0xffffffff, indicating
	/// that the movement has been executed as specified, or (newy<<16+newx), giving the nearest possible valid sprite position.
	/// \param obj The sprite that has moved.
	/// \param oldx The sprite's X position before the movement.
	/// \param oldy The sprite's Y position before the movement.
	/// \returns 0xfffffff or the best possible valid sprite position.
	virtual u32 spritemoved(sprite* obj, int oldx, int oldy);
	
private:
	/// Called by any sprite that pushes another one, i.e. that collides with the other sprite. The default behavior is to hand such
	/// an event on to the affected sprite. Extending this method enables you to perform special actions when two sprites collide.
	/// \param me The sprite that pushes.
	/// \param them The sprite that is pushed.
	virtual void push(sprite* me, sprite* them);

	/// Register a sprite with this stage. Adds a sprite to the stage's sprite list.
	/// \param sprite The sprite to be added.
	void addsprite(sprite* sprite);
	
	/// Removes a sprite from this stage's sprite list. This method does not delete the sprite altogether, it only removes it
	/// from the sprite list.
	/// \param obj The sprite to be removed.
	void deletesprite(sprite* obj);
	
	/// Checks the given coordinates for collision with the collision map. You give a set of box edges, checkcollide() returns whether
	/// this box overlaps with a non-zero block on the collision map. Coordinates are given in units of 8x8 tiles.
	/// \param collmap Collision map to be used.
	/// \param left Left edge.
	/// \param right Right edge.
	/// \param top Top edge.
	/// \param bottom Bottom edge.
	/// \returns Collision or not.
	bool checkcollide(int collmap, int left, int right, int top, int bottom) MEM_FUNC_IN_IWRAM;

	/// Switch to sprites in the sprite list in order to swap their priorities.
	/// \param first Number of first sprite.
	/// \param second Number of second sprite.
	void switchpriority(int first, int second);

	int m_lexcess;		///< Amount of map data to ignore at left border.
	int m_texcess;		///< Amount of map data to ignore at top border.
	int m_rexcess;		///< Amount of map data to ignore at right border.
	int m_bexcess;		///< Amount of map data to ignore at bottom border.

	int m_scrolllocked;	///< Number of sprites the stage is scroll locked on (-1: disabled).
	
	map_fragment_info_ptr* m_map[4];	///< Info structures for the maps.
	int m_first_bg;		///< Number of first HW background used.
	int m_num_bgs;		///< Number of backgrounds used

	int m_currentfrag[4];		///< Current map section for each of the four screen corners.
	int m_mapsectionsx;		///< Number of map sections in X direction.
	int m_mapsectionsy;		///< Number of map sections in Y direction.
	const u16* m_mapdata[4];		///< Map data in 32x32 chunks, optimized for DMA transfar.
	const u8* m_mapdata_unopt[4];	///< Linear representation of map data with flip bits clipped.
	bool m_mapinvert[4];		///< Collision detection inversion flags.
	BLKSTR* m_blockstr;			///< Block info from Mappy.
	sprite* m_sprites[MAX_OBJECTS];	///< The stage's sprite list.
	sprite* m_staticsprites[MAX_OBJECTS];	///< The static sprite list used by scriptable sprites to find each other
	int m_spriteprofile[MAX_OBJECTS];		///< Maximum time the sprite's compute() time takes.
	int m_fixedprio;		///< Number of sprites that have a fixed priority;
	int m_maxcomputetime;					///< Maximum time the compute() method may take, minus 10% for safety.
	int m_spritecnt;		///< Number of sprites registered with the stage.
	volatile unsigned int m_framecnt;	///< Frame counter.
	int m_scrollborderx;	///< Horizontal scroll locking border.
	int m_scrollbordery;	///< Vertical scroll locking border.

	sprite* m_sighted;		///< Sprite that was sighted.
	sprite* m_sightedby;	///< Sprite that has sighted.

	u32 m_signals;			///< State of the stage's signals during the current compute() round.
	u32 m_signalsnext;		///< State of the stage's signals after the current compute() round.

	int m_ticks;			///< Game time counter, incremented by compute()
	
	/// Draws the radar screen if enabled.
	void drawradar() MEM_FUNC_IN_IWRAM;

	int m_radarbg;					///< HW background used for radar screen
	bool m_radarenable;				///< Radar enabled?
	const palettedata* m_radarpalette;		///< 16-color palette to be used for radar
	int m_radarx;					///< X position of radar screen.
	int m_radary;					///< Y position of radar screen.
	int m_radarxsize;				///< Horizontal size of radar screen, in pixels.
	int m_radarysize;				///< Vertical size of radar screen, in pixels.

	static const palettedata m_radardefpal;	///< Default radar screen palette.

	int m_numhighlights;			///< Number of active highlights.
	sprite* m_highlightsprite[S_MAXHIGHLIGHTS];		///< The sprites used for the object window when highlighting.
	int m_highlightoldxpos;			///< The stage position before the first call to addhighlight(), X coord.
	int m_highlightoldypos;			///< The stage position before the first call to addhighlight(), Y coord.

	int m_defaultcolmap;			///< The default collide map.

	int m_numcamhints;				///< Number of registered camera hints.
	camhint m_camhints[S_MAXCAMHINTS];	///< Camera hints.
};

#endif
