// (C) 2003 Ulrich Hecht <uli@emulinks.de>.
// Distributed under the terms of the GNU Public License v2.

#ifndef _VBL_H
#define _VBL_H

extern "C" {
#include <mygba.h>
}

#include <object.h>

#define VBL_MAX_HANDLERS 10		///< The maximum number of interrupt handlers that can be registered.

/// Do every VBL_FRAMESKIPth frame. 1 means full framerate. Be sure to multiply all speed constants (except those in scripts) with
/// this value to ensure a constant game speed. This is generally assumed to be a power of two. Arguably the only useful values are
/// 1 (60 fps) and 2 (30 fps).
#define VBL_FRAMESKIP 2

/// VBlank interrupt handler class. This is a container class for functions to be called during VBlank. It automatically
/// enables and disables the VBlank interrupt depending on how many functions are registered with it. vbl is a singleton.
class vbl
{
public:
	/// Returns a pointer to the single instance of this class.
	/// \returns Pointer to the VBlank handler object.
	static vbl* instance();

	/// Register an object's interrupt handler method. The first registration will automatically enable the VBlank interrupt.
	/// \param obj The object who's vbl_handler() method should be called.
	void addhandler(object* obj);

	/// Deregister an object's interrupt handler method. If the given object is the last one in the list this will automatically
	/// disable the VBlank interrupt handler.
	/// \param obj The object to be removed.
	void removehandler(object* obj);

	/// Suspend the interrupt handling. When this is set to true the master handler will not call any of the
	/// handler functions registered. Setting it to false resumes normal operation. Use this when manipulating
	/// resources also used by VBlank handlers, e.g. DMA.
	/// \param susp Suspend interrupt handling?
	void suspend(bool susp) { m_vblsuspended=susp;}

	/// Returns the frame counter. Remember this counts real time as opposed to game time. For the latter refer to stage::getticks().
	/// Also note that the counter is suspended while the handler list is empty. It is independent of VBL_FRAMESKIP and always counts
	/// at 60 fps.
	/// \returns Current frame counter value.
	u32 getframecount() { return m_framecount;}
	
	/// Waits in low-power mode until the VBlank handler has set m_newframe to true.
	void waitfornextframe() { while(!m_newframe) { asm("swi 2"); } ; m_newframe=false; }

protected:
	/// The vbl constructor.
	vbl();
private:
	static void masterhandler();	///< The master VBlank interrupt handler.
	static vbl* m_instance;							///< The vbl instance

	object* m_handlerlist[VBL_MAX_HANDLERS];	///< The list of interrupt handlers.
	int m_handlernum;								///< The number of interrupt handlers registered.
	bool m_interruptenabled;						///< Whether the master handler is enabled.
	bool m_vblsuspended;							///< Interrupt suspended. Set this when modifying the handler list.

	u32 m_framecount;							///< The frame counter.
	volatile bool m_newframe;					///< The new frame flag.
};

// convenience macros
#define VBL_SUSPEND(x) vbl::instance()->suspend(x)
#define VBL_NEXT_FRAME vbl::instance()->waitfornextframe();

#endif

