// (C) 2003 Ulrich Hecht <uli@emulinks.de>.
// Distributed under the terms of the GNU Public License v2.

#ifndef _WIDGET_H
#define _WIDGET_H

#include <object.h>

#include <formats.h>

#define W_TOTAL_NUM_CHARS 224	///< Number of characters in bitmap. For console_font.bmp this is 224.
#define W_NUM_CHARS 96			///< Number of characters to be loaded into VRAM. For console_font.bmp this is 96 for all letters and 64 for capitals only.
#define W_CHAR_WIDTH 1			///< Width of character in tiles.
#define W_CHAR_HEIGHT 2			///< Height of character in tiles.
#define W_TILES_PER_CHAR (W_CHAR_WIDTH*W_CHAR_HEIGHT)		///< Number of 8x8 tiles per character.
#define W_TOTAL_NUM_CHAR_TILES (W_TOTAL_NUM_CHARS*W_TILES_PER_CHAR)	///< Total number of charset tiles in complete bitmap.
#define W_NUM_CHAR_TILES (W_NUM_CHARS*W_TILES_PER_CHAR)		///< Total number of charset tiles in used portion of bitmap.

#define W_CHAR_ASCII_OFFSET 32	///< ASCII character the bitmap starts at.
#define W_CHAR_BITMAP_WIDTH 16	///< Width of bitmap in characters.

/// The base class for GUI widgets. It allocates one HW background (two if using a backdrop) and the 16-color BG palette number 15,
/// i.e. the 16 topmost BG palette entries, as well as a number of tiles. When transparency is enabled, it also uses the HW special
/// effects. All positions are relative to the top left corner of the widget, regardless of border state, unless otherwise specified.
class widget : public object
{
public:
	/// Data type describing the tiles that make up the widget border. t_startofaddongfx is the start of the tileset area reserved
	/// for derived classes by the numtiles parameter of the widget constructor.
	enum decoration { t_tl=W_NUM_CHAR_TILES+1,t_top,t_tr,t_right,t_br,t_bottom,t_bl,t_left,t_empty,t_startofaddongfx };

	/// Widget constructor.
	/// \param bg The HW background to be used.
	/// \param border The tileset to be used for the border.
	/// \param numtiles The number of additional tiles used by derived classes.
	/// \param xsize The horizontal size of the widget, in 8x8 tiles.
	/// \param ysize The vertical size of the widget, in 8x8 tiles.
	/// \param xpos The X position of the widget, in pixels.
	/// \param ypos The Y position of the widget, in pixels.
	/// \param textdelay Number of frames to pause between the output of two characters.
	/// \param transparent Make the widget transparent.
	widget(int bg, const spritedata* border, int numtiles,int xsize,int ysize,int xpos=0,int ypos=0, int textdelay=0, bool transparent=false);

	/// The widget destructor.
	virtual ~widget();

	/// Sets the position of the widget
	/// \param x X coordinate of top left corner, in pixels, relative to the visible screen.
	/// \param y Y coordinate of top left corner, in pixels, relative to the visible screen.
	virtual void setpos(int x, int y);

	/// Clears the widget, keeping the border if there is one.
	void clear();
	
	/// Draws an ASCII character at the specified position.
	/// \param x X position of the character, in tiles.
	/// \param y Y position of the character, in tiles.
	/// \param c An ASCII character.
	void drawasciichar(int x, int y, unsigned char c);
	
	/// Draws a 0-terminated ASCII string at the specified position.
	/// \param x X position of the string, in tiles.
	/// \param y Y position of the string, in tiles.
	/// \param string The text string.
	void drawstring(int x, int y, const char* string);

	/// Draws num characters from an ASCII string at the specified position. Note that this method does not
	/// check for end-of-string characters.
	/// \param x X position of the string, in tiles.
	/// \param y Y position of the string, in tiles.
	/// \param string The text string.
	/// \param num The number of characters to be drawn.
	void drawstringn(int x, int y, const char* string, int num);
	
	/// Sets the character output delay. Setting this to n causes the string and character output functions to
	/// pause n frames between each character.
	/// \param delay Character write delay in frames.
	void settextdelay(int delay) { m_textdelay=delay;}

	/// Returns the character output delay.
	/// \returns Character write delay in frames.
	int gettextdelay() { return m_textdelay;}

	/// Use up another background to display a background image behind the main widget.
	/// \param backdrop Backdrop graphics.
	void setbackgroundimage(const backdropdata* backdrop);

	/// Sets the tileset used for text output. Expects LZ77-compressed data. The size of this tileset is determinded by the
	/// W_NUM_CHARS and W_TILES_PER_CHAR defines.
	/// \param tileset The LZ77-compressed character set tileset.
	static void setcharacters(const fontdata* tileset) { m_characters=tileset; }

	/// Sets the combined 16-color palette for the border and the character set.
	/// \param palette A 16-color palette.
	static void setpalette(const palettedata* palette) { m_palette=palette; }

	/// Widgets are not part of the game world. Therefore moveme() must be called in the vbl_handler(), not in compute().
	virtual void compute() {}

	virtual void vbl_handler();

protected:
	int m_bg;	///< The HW background used by the widget.
	int m_borderwidth;	///< Width of border in tiles.
	
private:
	const spritedata* m_border;	///< The tileset used for the border.
	
	int m_textdelay;	///< Number of frames to pause between character writes.

	bool m_havebgimage;	///< Do we have a background image?

	static const fontdata* m_characters;	///< The tileset used for the character set.
	static const palettedata* m_palette;		///< The 16-color palette used for both the charset and the border.

};

#endif

