/*
** $Id: ldo.c,v 1.109a 2000/10/30 12:38:50 roberto Exp $
** Stack and Call structure of Lua
** See Copyright Notice in lua.h
*/


#include <setjmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lua.h"

#include "ldebug.h"
#include "ldo.h"
#include "lgc.h"
#include "lmem.h"
#include "lobject.h"
#include "lparser.h"
#include "lstate.h"
#include "lstring.h"
#include "ltable.h"
#include "ltm.h"
#include "lundump.h"
#include "lvm.h"
#include "lzio.h"


/* space to handle stack overflow errors */
#define EXTRA_STACK	(2*LUA_MINSTACK)


void luaD_init (lua_State *L, int stacksize) {
  L->stack = luaM_newvector(L, stacksize+EXTRA_STACK, TObject);
  L->nblocks += stacksize*sizeof(TObject);
  L->stack_last = L->stack+(stacksize-1);
  L->stacksize = stacksize;
  L->Cbase = L->top = L->stack;
}


void luaD_checkstack (lua_State *L, int n) {
  if (L->stack_last - L->top <= n) {  /* stack overflow? */
    if (L->stack_last-L->stack > (L->stacksize-1)) {
      /* overflow while handling overflow */
      luaD_breakrun(L, LUA_ERRERR);  /* break run without error message */
    }
    else {
      L->stack_last += EXTRA_STACK;  /* to be used by error message */
      lua_error(L, "stack overflow");
    }
  }
}


static void restore_stack_limit (lua_State *L) {
  if (L->top - L->stack < L->stacksize - 1)
    L->stack_last = L->stack + (L->stacksize-1);
}


/*
** Adjust stack. Set top to base+extra, pushing NILs if needed.
** (we cannot add base+extra unless we are sure it fits in the stack;
**  otherwise the result of such operation on pointers is undefined)
*/
void luaD_adjusttop (lua_State *L, StkId base, int extra) {
  int diff = extra-(L->top-base);
  if (diff <= 0)
    L->top = base+extra;
  else {
    luaD_checkstack(L, diff);
    while (diff--)
      ttype(L->top++) = LUA_TNIL;
  }
}


/*
** Open a hole inside the stack at `pos'
*/
static void luaD_openstack (lua_State *L, StkId pos) {
  int i = L->top-pos; 
  while (i--) pos[i+1] = pos[i];
  incr_top;
}


static void dohook (lua_State *L, lua_Debug *ar, lua_Hook hook) {
  StkId old_Cbase = L->Cbase;
  StkId old_top = L->Cbase = L->top;
  luaD_checkstack(L, LUA_MINSTACK);  /* ensure minimum stack size */
  L->allowhooks = 0;  /* cannot call hooks inside a hook */
  (*hook)(L, ar);
  LUA_ASSERT(L->allowhooks == 0, "invalid allow");
  L->allowhooks = 1;
  L->top = old_top;
  L->Cbase = old_Cbase;
}


void luaD_lineHook (lua_State *L, StkId func, int line, lua_Hook linehook) {
  if (L->allowhooks) {
    lua_Debug ar;
    ar._func = func;
    ar.event = "line";
    ar.currentline = line;
    dohook(L, &ar, linehook);
  }
}


static void luaD_callHook (lua_State *L, StkId func, lua_Hook callhook,
                    const char *event) {
  if (L->allowhooks) {
    lua_Debug ar;
    ar._func = func;
    ar.event = event;
    infovalue(func)->pc = NULL;  /* function is not active */
    dohook(L, &ar, callhook);
  }
}


static StkId callCclosure (lua_State *L, const struct Closure *cl, StkId base) {
  int nup = cl->nupvalues;  /* number of upvalues */
  StkId old_Cbase = L->Cbase;
  int n;
  L->Cbase = base;       /* new base for C function */
  luaD_checkstack(L, nup+LUA_MINSTACK);  /* ensure minimum stack size */
  for (n=0; n<nup; n++)  /* copy upvalues as extra arguments */
    *(L->top++) = cl->upvalue[n];
  n = (*cl->f.c)(L);  /* do the actual call */
  L->Cbase = old_Cbase;  /* restore old C base */
  return L->top - n;  /* return index of first result */
}


void luaD_callTM (lua_State *L, Closure *f, int nParams, int nResults) {
  StkId base = L->top - nParams;
  luaD_openstack(L, base);
  clvalue(base) = f;
  ttype(base) = LUA_TFUNCTION;
  luaD_call(L, base, nResults);
}


/*
** Call a function (C or Lua). The function to be called is at *func.
** The arguments are on the stack, right after the function.
** When returns, the results are on the stack, starting at the original
** function position.
** The number of results is nResults, unless nResults=LUA_MULTRET.
*/ 
void luaD_call (lua_State *L, StkId func, int nResults) {
  L->callcount++;	/* yield patch: recursion protection for yield() */
  if (luaD_prolog(L, func, nResults) == 0) {
    L->callinfostack->mustreturn = 1; /* yield patch: tell VM this frame must return */
    luaV_execute(L);
  }
  L->callcount--;
}

/*
** yield patch: Push a fresh CallInfo structure onto L->callinfostack.
*/
static void pushcallinfo (lua_State* L)
{
  CallInfo* ci = L->callinfofreelist;
  if (ci == NULL) {
    /* allocate a new CallInfo struct. */
    ci = luaM_malloc(L, sizeof(CallInfo));
  } else {
    L->callinfofreelist = ci->prev;
  }
  ci->prev = L->callinfostack;
  L->callinfostack = ci;
}

/*
** yield patch: Pop the top CallInfo struct off L->callinfostack.
*/
static void popcallinfo (lua_State* L)
{
  CallInfo* ci = L->callinfostack;
  LUA_ASSERT(ci, "popcallinfo(): callinfo stack is empty");
  if (ci) {
    L->callinfostack = ci->prev;

    /* add the popped struct to the callinfofreelist, so it can be reused easily. */
    ci->prev = L->callinfofreelist;
    L->callinfofreelist = ci;
  }
}


static void luaV_pack (lua_State *L, StkId firstelem) {
  int i;
  Hash *htab = luaH_new(L, 0);
  for (i=0; firstelem+i<L->top; i++)
    *luaH_setint(L, htab, i+1) = *(firstelem+i);
  /* store counter in field `n' */
  luaH_setstrnum(L, htab, luaS_new(L, "n"), i);
  L->top = firstelem;  /* remove elements from the stack */
  ttype(L->top) = LUA_TTABLE;
  hvalue(L->top) = htab;
  incr_top;
}


static void adjust_varargs (lua_State *L, StkId base, int nfixargs) {
  int nvararg = (L->top-base) - nfixargs;
  if (nvararg < 0)
    luaD_adjusttop(L, base, nfixargs);
  luaV_pack(L, base+nfixargs);
}


/*
** yield patch: Function-call prolog.  Function is at func, args are
** between (func,top).  Initialize the given CallInfo structure, to
** use for activating luaV_execute().
**
** Returns 1 if we handled the call directly in this function --
** happens when the function turns out to be a C function.  Caller
** _should not_ follow this with luaV_execute().  Returns 0 if
** luaV_execute should be called.
*/
int luaD_prolog (lua_State *L, StkId func, int nResults) {
  CallInfo* ci;
  Closure *cl;

  pushcallinfo(L);
  ci = L->callinfostack;
  ci->funcid = func;
  ci->nResults = nResults;
  ci->mustreturn = 0;

  if (ttype(func) != LUA_TFUNCTION) {
    /* `func' is not a function; check the `function' tag method */
    Closure *tm = luaT_gettmbyObj(L, func, TM_FUNCTION);
    if (tm == NULL)
      luaG_typeerror(L, func, "call");
    luaD_openstack(L, func);
    clvalue(func) = tm;  /* tag method is the new function to be called */
    ttype(func) = LUA_TFUNCTION;
  }
  cl = clvalue(func);
  ci->func = cl;
  infovalue(func) = ci;
  ttype(func) = LUA_TMARK;
  ci->callhook = L->callhook;
  if (ci->callhook)
    luaD_callHook(L, func, ci->callhook, "call");

  if (ci->func->isC) {
    /* go ahead and call the C function.  Could cause recursion into the VM. */
    ci->savedpc = NULL;
    ci->firstResult = callCclosure(L, ci->func, func+1);
    luaD_epilog(L);
    return 1;
  }
  else {
    /* Lua function.  Set up the callinfo, for entering the VM. */

    /* yield patch: from the start of luaV_execute */
    const Proto *const tf = ci->func->f.l;
    ci->linehook = L->linehook;
    ci->savedpc = tf->code;
    ci->pc = &ci->savedpc;
    ci->base = func+1;
    luaD_checkstack(L, tf->maxstacksize+EXTRA_STACK);
    if (tf->is_vararg)  /* varargs? */
      adjust_varargs(L, ci->base, tf->numparams);
    else
      luaD_adjusttop(L, ci->base, tf->numparams);
    /* now we're ready to call luaV_execute */
    return 0;
  }
}


/*
** yield patch: cleanup and return from a function call.  Do callhook, and
** copy function results to the proper place on the stack.  Pop
** the callinfostack.
*/
void luaD_epilog (lua_State *L) {
  CallInfo* ci = L->callinfostack;
  LUA_ASSERT(ci, "no valid CallInfo for luaD_epilog");
  if (ci->callhook)  /* same hook that was active at entry */
    luaD_callHook(L, ci->funcid, ci->callhook, "return");
  LUA_ASSERT(ttype(ci->funcid) == LUA_TMARK, "invalid tag");

  /* move results to `funcid' (to erase parameters and function) */
  if (ci->nResults == LUA_MULTRET) {
	StkId func = ci->funcid;
    while (ci->firstResult < L->top)  /* copy all results */
      *func++ = *(ci->firstResult)++;
	L->top = func;
  }
  else {  /* copy at most `nResults' */
	StkId func = ci->funcid;
    for (; ci->nResults > 0 && ci->firstResult < L->top; ci->nResults--)
      *func++ = *ci->firstResult++;
    L->top = func;
    for (; ci->nResults > 0; ci->nResults--) {  /* if there are not enough results */
      ttype(L->top) = LUA_TNIL;  /* adjust the stack */
      incr_top;  /* must check stack space */
    }
  }
  luaC_checkGC(L);

  popcallinfo(L);        /* L->callinfostack = L->callinfostack->prev; and also save the old callinfo on the callinfofreelist */
}


/*
** Execute a protected call.
*/
struct CallS {  /* data to `f_call' */
  StkId func;
  int nresults;
};

static void f_call (lua_State *L, void *ud) {
  struct CallS *c = (struct CallS *)ud;
  luaD_call(L, c->func, c->nresults);
}


LUA_API int lua_call (lua_State *L, int nargs, int nresults) {
  StkId func = L->top - (nargs+1);  /* function to be called */
  struct CallS c;
  int status;
  /* yield patch: remove stale callinfo, since the caller didn't call lua_resume() on this lua_State */
  /* if (L->callcount == 0) { */
#if 0	/* REMEMBERME: Disabling this allows us to make interrupts with lua_call() and resume afterwards. */
    while (L->callinfostack)
      popcallinfo(L);
#endif
/*  } */
  c.func = func; c.nresults = nresults;
  status = luaD_runprotected(L, f_call, &c);
  if (status != 0 && status != LUA_YIELD) { /* an error occurred? */ /* yield patch: change to accommodate LUA_YIELD, was (status != 0) */
    L->top = func;  /* remove parameters from the stack */
  }

  return status;
}


/*
** Execute a protected parser.
*/
struct ParserS {  /* data to `f_parser' */
  ZIO *z;
  int bin;
};

static void f_parser (lua_State *L, void *ud) {
  struct ParserS *p = (struct ParserS *)ud;
  Proto *tf = p->bin ? luaU_undump(L, p->z) : luaY_parser(L, p->z);
  luaV_Lclosure(L, tf, 0);
}


static int protectedparser (lua_State *L, ZIO *z, int bin) {
  struct ParserS p;
  unsigned long old_blocks;
  int status;
  p.z = z; p.bin = bin;
  /* before parsing, give a (good) chance to GC */
  if (L->nblocks/8 >= L->GCthreshold/10)
    luaC_collectgarbage(L);
  old_blocks = L->nblocks;
  status = luaD_runprotected(L, f_parser, &p);
  if (status == 0) {
    /* add new memory to threshold (as it probably will stay) */
    L->GCthreshold += (L->nblocks - old_blocks);
  }
  else if (status == LUA_ERRRUN)  /* an error occurred: correct error code */
    status = LUA_ERRSYNTAX;
  return status;
}


static int parse_file (lua_State *L, const char *filename) {
  ZIO z;
  int status;
  int bin;  /* flag for file mode */
  int c;    /* look ahead char */
  FILE *f = (filename == NULL) ? stdin : fopen(filename, "r");
  if (f == NULL) return LUA_ERRFILE;  /* unable to open file */
  c = fgetc(f);
  ungetc(c, f);
  bin = (c == ID_CHUNK);
  if (bin && f != stdin) {
    f = freopen(filename, "rb", f);  /* set binary mode */
    if (f == NULL) return LUA_ERRFILE;  /* unable to reopen file */
  }
  lua_pushstring(L, "@");
  lua_pushstring(L, (filename == NULL) ? "(stdin)" : filename);
  lua_concat(L, 2);
  c = lua_gettop(L);
  filename = lua_tostring(L, c);  /* filename = '@'..filename */
  luaZ_Fopen(&z, f, filename);
  status = protectedparser(L, &z, bin);
  lua_remove(L, c);  /* remove `filename' from the stack */
  if (f != stdin)
    fclose(f);
  return status;
}


LUA_API int lua_dofile (lua_State *L, const char *filename) {
  int status = parse_file(L, filename);
  if (status == 0)  /* parse OK? */
    status = lua_call(L, 0, LUA_MULTRET);  /* call main */
  return status;
}


static int parse_buffer (lua_State *L, const char *buff, size_t size,
                         const char *name) {
  ZIO z;
  if (!name) name = "?";
  luaZ_mopen(&z, buff, size, name);
  return protectedparser(L, &z, buff[0]==ID_CHUNK);
}


LUA_API int lua_dobuffer (lua_State *L, const char *buff, size_t size, const char *name) {
  int status = parse_buffer(L, buff, size, name);
  if (status == 0)  /* parse OK? */
    status = lua_call(L, 0, LUA_MULTRET);  /* call main */
  return status;
}


LUA_API int lua_dostring (lua_State *L, const char *str) {
  return lua_dobuffer(L, str, strlen(str), str);
}


/* yield patch */
static void f_resume (lua_State* L, void* dummy) {
  if (L->callinfostack == NULL)
    lua_error(L, "invalid lua_resume()");	/* L was not exited via "yield()" */
  L->callcount = 1;	/* simulate the initial call to luaD_call that happens with lua_do* */
  luaV_execute(L);
  LUA_ASSERT(L->callcount == 1, "invalid callcount in f_resume");
  L->callcount = 0;
}


LUA_API int lua_resume (lua_State *L) {
  int status = luaD_runprotected(L, f_resume, NULL);
  return status;
}


/*
** {======================================================
** Error-recover functions (based on long jumps)
** =======================================================
*/

/* chain list of long jump buffers */
struct lua_longjmp {
  jmp_buf b;
  struct lua_longjmp *previous;
  volatile int status;  /* error code */
};


static void message (lua_State *L, const char *s) {
  const TObject *em = luaH_getglobal(L, LUA_ERRORMESSAGE);
  if (ttype(em) == LUA_TFUNCTION) {
    *L->top = *em;
    incr_top;
    lua_pushstring(L, s);
    luaD_call(L, L->top-2, 0);
  }
}


/*
** Reports an error, and jumps up to the available recovery label
*/
LUA_API void lua_error (lua_State *L, const char *s) {
  if (s) message(L, s);
  luaD_breakrun(L, LUA_ERRRUN);
}


void luaD_breakrun (lua_State *L, int errcode) {
  if (L->errorJmp) {
    L->errorJmp->status = errcode;
    longjmp(L->errorJmp->b, 1);
  }
  else {
    if (errcode != LUA_ERRMEM)
      message(L, "unable to recover; exiting\n");
    exit(EXIT_FAILURE);
  }
}


int luaD_runprotected (lua_State *L, void (*f)(lua_State *, void *), void *ud) {
  CallInfo* baseci = L->callinfostack; /* yield patch */
  int oldcallcount = L->callcount;	/* yield patch */
  StkId oldCbase = L->Cbase;
  StkId oldtop = L->top;
  struct lua_longjmp lj;
  int allowhooks = L->allowhooks;
  /* yield patch */
  if (L->callinfostack) {
    /* Resuming; use stashed values for allowhooks, oldCbase, oldtop. */
    LUA_ASSERT(L->yield_oldCbase, "yield patch: problem in luaD_runprotected.");
    LUA_ASSERT(L->yield_oldtop, "yield patch: problem in luaD_runprotected.");
    oldCbase = L->yield_oldCbase;
    oldtop = L->yield_oldtop;
    allowhooks = L->yield_allowhooks;
  }
  lj.status = 0;
  lj.previous = L->errorJmp;  /* chain new error handler */
  L->errorJmp = &lj;
  if (setjmp(lj.b) == 0)
    (*f)(L, ud);
  else if (lj.status == LUA_YIELD) {
    /* yield patch: Script yielded. */
    LUA_ASSERT(L->callinfostack, "yield patch: runprotected exited with LUA_YIELD, but callinfostack is 0!");
    /* Save the allowhooks, oldCbase, and oldtop values, for restoration
       by a future exit from lua_resume(). */
    L->yield_oldCbase = oldCbase;
    L->yield_oldtop = oldtop;
    L->yield_allowhooks = allowhooks;
  }
  else if (lj.status != LUA_YIELD) {  /* an error occurred: restore the state */  /* yield patch: don't cleanup if we're about to yield! */
    while (L->callinfostack && L->callinfostack != baseci)  /* yield patch: unwind the callinfo stack, to the point where we started */
      popcallinfo(L);
    L->callcount = oldcallcount;  /* yield patch */
    /* yield patch: What if this fn was called from lua_resume?
       These restores will be wrong.

       TODO: restore to the values of the original run_protected, that
       was yielded out of. */
    L->allowhooks = allowhooks;
    L->Cbase = oldCbase;
    L->top = oldtop;
    restore_stack_limit(L);
  }
  L->errorJmp = lj.previous;  /* restore old error handler */
  return lj.status;
}

/* }====================================================== */

LUA_API int lua_yield (lua_State *L) {
  if (L->callcount == 1) {	/* one original call to luaD_call, which we will unwind via longjmp(). */
    L->callcount = 0;

    /* unwind L's stack to remove this invocation of lua_yield */
    L->top = L->Cbase - 1;
    L->Cbase = L->stack;
    popcallinfo(L);

    /* non-local exit, to original C caller */
    luaD_breakrun(L, LUA_YIELD);
  }
  else {
    /* can't unwind through C functions. */
    /* Print warning message, and don't yield. */
    message(L, "lua_yield: Can't yield from inside recursive call to luaD_call!");
  }
  return 0;
}
