// (C) 2003 Ulrich Hecht <uli@emulinks.de>.
// Distributed under the terms of the GNU Public License v2.

#include <sprite.h>
#include <stage.h>
#include <vbl.h>
#include <misc.h>
#include <hardware.h>

sprite::sprite(stage* parent, const spritedata* gfx, int xpos, int ypos)
{
	u32 shape=0xff,size=0xff;	// init to undefined value

	m_xpos=xpos;
	m_ypos=ypos;
	m_xsize=gfx->xsize;
	m_ysize=gfx->ysize;

	m_movedeltax=m_movedeltay=m_movecountdown=0;
	m_ismoving=false;
	m_path=0;
	m_ispathing=false;
	m_pathstep=0;
	m_nodecounter=0;
	m_maxnodecount=1024; // one node is 16 bytes the last time I checked
	m_maxtime=35*11*VBL_FRAMESKIP;	// 35 percent per VBL_FRAMESKIP frames for A*
	m_maxtimepercent=35;
	m_resumepathing=false;
	m_open=0;
	m_closed=0;
	m_deleteopen=m_deleteclosed=false;
	m_pathxdest=0;
	m_pathydest=0;
	m_honorsprites=false;

	m_movetype=t_straight;

	m_animations=gfx->animations;
	m_currentanimation=m_lastanimation=0;
	m_currentdirection=d_left;	// equals 0
	m_phases=gfx->phases;
	m_currentphase=m_lastphase=0;
	m_eventsperphase=m_eventcount=gfx->eventsperphase;
	m_animtype=a_timed;
	m_visible=true;
	m_screenrelative=false;

	m_collide=m_sightable=m_audible=false;
	m_hasmoved=m_hasmovedlastturn=false;
	m_collidemap=0;
	m_spritecollide=false;
	m_blockssight=false;
	m_facetolerance=4;
	m_adjust=12;
	m_adjustspeed=1;
	m_bounding[b_left]=m_bounding[b_top]=m_bounding[b_right]=m_bounding[b_bottom]=0;
	m_graphics=gfx->graphics;
	m_parent=parent;
	m_colormode=gfx->col_mode;

	m_radarcolor=0;

	m_rotatable=false;
	m_turning=false;
	m_turntoangle=0;
	m_turnspeednum=m_turnspeednumdeg=0;
	m_turnspeeddenom=0;
	m_turncountdown=0;

	m_angle=m_angledeg=0;
	m_repath=false;
	m_prevx=m_prevy=0;
	m_rotset=0;

	m_shovable=false;

	m_description="A sprite.";

	ASSERT(m_xsize==8||m_xsize==16||m_xsize==32||m_xsize==64,"invalid x dimension");
	ASSERT(m_ysize==8||m_ysize==16||m_ysize==32||m_ysize==64,"invalid y dimension");
	if(m_xsize==8)
	{
		switch(m_ysize)
		{
		case 8: shape=0; size=0; break;
		case 16: shape=2; size=0; break;
		case 32: shape=2; size=1; break;
		default: ASSERT(0,"invalid combination of dimensions");
		}
	}
	else if(m_xsize==16)
	{
		switch(m_ysize)
		{
		case 8: shape=1; size=0; break;
		case 16: shape=0; size=1; break;
		case 32: shape=2; size=2; break;
		default: ASSERT(0,"invalid combination of dimensions");
		}
	}
	else if(m_xsize==32)
	{
		switch(m_ysize)
		{
		case 8: shape=1; size=1; break;
		case 16: shape=1; size=2; break;
		case 32: shape=0; size=2; break;
		case 64: shape=2; size=3; break;
		}
	}
	else if(m_xsize==64)
	{
		switch(m_ysize)
		{
		case 32: shape=1; size=3; break;
		case 64: shape=0; size=3; break;
		default: ASSERT(0,"invalid combination of dimensions");
		}
	}

	VBL_SUSPEND(true);
	m_objno=ham_CreateObj((void*)(gfx->graphics),shape,size,OBJ_MODE_NORMAL,
						  gfx->col_mode,	// 256 colors
						  gfx->palette, // unused with 256 colors
						  0,0,0, // mosaic/hflip/vflip
						  1, // priority
						  0, // double size
						  0,0);
	VBL_SUSPEND(false);

	// ham_CreateObj() does not accept out of bounds positions, while ham_SetObjXY() does.
	ham_SetObjXY(m_objno, xpos-parent->getxpos(),ypos-parent->getypos());

	parent->addsprite(this);
}

//#define DEBUG_SPRITEDTOR

sprite::~sprite()
{
	ham_DeleteObj(m_objno);
	m_parent->deletesprite(this);
	if(m_path) delete[] m_path;
	if(m_rotatable)	freerotset(m_rotset);
	
#ifdef DEBUG_SPRITEDTOR
	uprintf("sprite %s at 0x%x deleted\n", getdescription(), (int)this);
#endif

	// A* queue deletion is done when deleting stage. This is, admittedly, a strange approach, but it avoids
	// load peaks at inappropriate times.
}

//#define DEBUG_ANIMATION

void sprite::redraw()
{
#ifdef DEBUG_ANIMATION
	uprintf("ph %d phss %d\n",phase,phases);
#endif
	if (m_eventcount <= 0)
	{
		m_eventcount=m_eventsperphase;
		if (++m_currentphase == m_phases)
			m_currentphase=0;
	}
#ifdef DEBUG_ANIMATION
	uprintf("fr %d fpp %d ph %d\n",frame,framesperphase,phase);
#endif

	if (m_currentanimation!=m_lastanimation || m_currentphase!=m_lastphase)
	{
		ham_UpdateObjGfx(m_objno,(char*)m_graphics+(m_currentanimation*m_phases+m_currentphase)*m_xsize*(m_ysize>>(-(m_colormode-1))));
		m_lastanimation=m_currentanimation; m_lastphase=m_currentphase;
	}
}

void sprite::vbl_handler()
{
	if(m_animtype == a_timed)
		m_eventcount-=VBL_FRAMESKIP;

	redraw();
}

//#define PROFILE_SETPOS
//#define DEBUG_SETPOS

//#define CHECK_MAP_BOUNDARIES

void sprite::setpos(int x, int y)
{
	int oldx,oldy;
	u32 newpos;

	if(m_audible)
		if(m_xpos!=x || m_ypos!=y) m_hasmoved=true;

#ifdef CHECK_MAP_BOUNDARIES
	if(x<-m_bounding[b_left]) x=-m_bounding[b_left];
	if(y<-m_bounding[b_top]) y=-m_bounding[b_top];
	if(x>m_parent->getxsize()*8-getboundxsize()) x=m_parent->getxsize()*8-getboundxsize();
	if(y>m_parent->getysize()*8-getboundysize()) y=m_parent->getysize()*8-getboundysize();
#endif

	oldx=getxpos();
	oldy=getypos();
	m_xpos=x;
	m_ypos=y;

	// check for sprite collision
	if(m_spritecollide)
	{
		for(int i=0;i<m_parent->m_spritecnt;i++)
		{
			sprite* spr=m_parent->m_sprites[i];
			if(spr==this) continue;
			if(collides(spr))
			{
				int dx=SGN(m_xpos-oldx);
				int dy=SGN(m_ypos-oldy);
				// try to shorten the movement until it does not collide any more
				while((!dx || m_xpos!=oldx) && (!dy || m_ypos!=oldy))
				{
					m_xpos-=dx; m_ypos-=dy;
					if(!collides(spr))
						break;
				}
				m_parent->push(this,spr);
			}
		}
	}
	
	// check for bg collision
#ifdef PROFILE_SETPOS
	int vcnt=R_VCNT;
#endif
	if((newpos=m_parent->spritemoved(this,oldx+m_bounding[b_left],oldy+m_bounding[b_top]))!=0xffffffff)
	{
		m_xpos=(newpos&0xffff)-m_bounding[b_left];
		m_ypos=(newpos>>16)-m_bounding[b_top];
		m_parent->spritemoved(this,0x7fff,0x7fff);
	}
#ifdef PROFILE_SETPOS
	uprintf("objm vcntb %d vcnta %d\n",vcnt,R_VCNT);
#endif

	if(m_animtype == a_motion)
	{
		m_eventcount-=abs(oldx-m_xpos)+abs(oldy-m_ypos);
#ifdef DEBUG_SETPOS
		uprintf("eventcount %d delta %d %d\n", m_eventcount,abs(oldx-m_xpos),abs(oldy-m_ypos));
#endif
	}

	reposition();
}

void sprite::reposition()
{
	int spritex,spritey;
	spritex=m_xpos-(m_screenrelative?0:m_parent->getxpos());
	spritey=m_ypos-(m_screenrelative?0:m_parent->getypos());

	// rot sprites are double size and thus need some position tweaking.
	if(m_rotatable)
	{
		spritex-=m_xsize/2;
		spritey-=m_ysize/2;
	}

	// park sprites outside visible screen to avoid wraparounds
	if(spritex>256 || spritex<-m_xsize)	spritex=256;
	if(spritey>196 || spritey<-m_ysize)	spritey=196;

	ham_SetObjXY(m_objno, spritex, spritey);
}

sprite* sprite::faces()
{
	sprite** sprites=m_parent->m_sprites;
	int num=m_parent->m_spritecnt;
	int sx,sy,sxs,sys;
	int mx,my,mxs,mys;

	ASSERT(m_currentdirection<=d_down,"unknown direction");

	mx=getxpos();
	my=getypos();
	mxs=getxsize();
	mys=getysize();

	for(int i=0;i<num;i++)
	{
		if(sprites[i]==this || (!sprites[i]->isvisible())) continue;

		sx=sprites[i]->getxpos();
		sy=sprites[i]->getypos();
		sxs=sprites[i]->getxsize();
		sys=sprites[i]->getysize();

		switch(m_currentdirection)
		{
		case d_up:
			if(sx<mx+mxs/2 && sx+sxs > mx+mxs/2 &&
			   sy<my && sy>my-sys-m_facetolerance)
				return sprites[i];
			break;
		case d_down:
			if(sx<mx+mxs/2 && sx+sxs > mx+mxs/2 &&
			   sy+sys>my+mys && sy<my+mys+m_facetolerance)
				return sprites[i];
			break;
		case d_left:
			if(sx<mx && sx+sxs>mx-m_facetolerance &&
			   sy<my+mys/2 && sy+sys>my+mys/2)
				return sprites[i];
			break;
		case d_right:
			if(sx<mx+mxs+m_facetolerance && sx+sxs>mx+mxs &&
			   sy<my+mys/2 && sy+sys>my+mys/2)
				return sprites[i];
			break;
		default:
			ASSERT(0, "this should never be reached");
			break;
		}
	}
	return NULL;
}

void sprite::setcollidemapno(int map)
{
	ASSERT(map<m_parent->m_num_bgs,"cannot use undefined map for collision");
	m_collidemap=map;
	ham_SetObjPrio(m_objno,3-(m_parent->m_first_bg+map));
}

//#define PROFILE_ASTAR
//#define DEBUG_ASTAR

#ifdef PROFILE_ASTAR
TOOL_PROFILER_INIT;
#endif

void sprite::moveto(int x, int y, int speed_num, int speed_denom)
{
	if(m_movetype==t_astar)
	{
		if(m_deleteopen) {
			delete m_open; m_deleteopen=false;
		}
		if(m_deleteclosed) {
			delete m_closed; m_deleteclosed=false;
		}

		if(m_path)
		{
			delete[] m_path;
			m_path=0;
			m_ispathing=false;
		}
#ifdef PROFILE_ASTAR
		TOOL_PROFILER_START;
#endif
		m_path=findpath(m_xpos/8,m_ypos/8,x/8,y/8);
#ifdef PROFILE_ASTAR
		TOOL_PROFILER_END;
#endif

#ifdef DEBUG_ASTAR
		if(m_path)
		{
			uprintf("from %d/%d to %d/%d\n", m_xpos/8, m_ypos/8, x/8, y/8);
			int count=0;
			while(m_path[count*2]!=0xffff)
			{
				uprintf("step %d: %d %d\n",count,m_path[count*2],m_path[count*2+1]);
				count++;
			}
		}
#endif

		if(m_path)
		{
			m_ispathing=true;
			m_pathstep=0;
		}
		m_movespeed_num=speed_num;
		m_movespeed_denom=speed_denom;

		if(!m_resumepathing) m_honorsprites=false;	// We're finished, next run is normal again.
	}
	else
		object::moveto(x-m_bounding[b_left], y-m_bounding[b_top], speed_num, speed_denom);
}

//#define DEBUG_PATHWALKING

void sprite::nextstepinpath()
{
	if(m_path[m_pathstep*2]==0xffff)   // end of path
	{
		m_ispathing=false;
		delete[] m_path;
		m_path=0;
		return;
	}

	if(m_pathstep && moveblocked())	// could not complete last step, requesting a repathing
	{
#ifdef DEBUG_PATHWALKING
		uprintf("blocked at step %d (%d/%d)\n",m_pathstep-1,m_path[(m_pathstep-1)*2],m_path[(m_pathstep-1)*2+1]);
		if(faces()) uprintf("facing %s at %d %d\n", faces()->getdescription(), faces()->getboundxpos(), faces()->getboundypos());
#endif

		// Backtracking. This is necessary to avoid problems caused by the implicit rounding involved in pathfinding on the map grid.
		// It also looks halfway realistic.
		object::moveto(m_prevx, m_prevy, m_movespeed_num, m_movespeed_denom);

		m_repath=true;
		return;
	}

	int x=m_path[m_pathstep*2]*8;
	int y=m_path[m_pathstep*2+1]*8;

	// find the most suitable animdirection
	if(abs(x-m_xpos)>abs(y-m_ypos))
	{
		if(x<m_xpos) setanimdirection(d_left);
		if(x>m_xpos) setanimdirection(d_right);
	}
	else
	{
		if(y<m_ypos) setanimdirection(d_up);
		if(y>m_ypos) setanimdirection(d_down);
	}

	// sprite::setpos() ignores bounding boxes, thus this whole shebang...
	m_prevx=getxpos(); m_prevy=getypos();
	object::moveto(x+m_xsize/2-getboundxsize()/2-m_bounding[b_left], y+m_ysize/2-getboundysize()/2-m_bounding[b_top], m_movespeed_num, m_movespeed_denom);
	m_pathstep++;
}

//#define PROFILE_SPRITE
//#define DEBUG_TURNING

#ifdef PROFILE_SPRITE
TOOL_PROFILER_INIT;
#endif

void sprite::compute(bool lowtime)
{
#ifdef PROFILE_SPRITE
	TOOL_PROFILER_START;
#endif
	
	// do the homework first
	if(m_deleteopen)
	{
		delete m_open;
		m_deleteopen=false;
	}
	if(m_deleteclosed)
	{
		delete m_closed;
		m_deleteclosed=false;
	}

	// If this is an A*-enabled sprite, either do queue deletion or, if nothing to delete, pathing.
	// May take a while, so skip it if we're low on time.
	if(m_movetype==t_astar)
	{
		if(m_resumepathing && !lowtime)	// A* pathing needs to be resumed
			moveto(0, 0, m_movespeed_num, m_movespeed_denom);

		// There is a path, we are following it, and the last step we took is through.
		if(m_path && m_ispathing && !m_ismoving)
		{
			if(m_repath)	// We hit something, most likely a sprite. Repath.
			{
				m_repath=false;
				m_honorsprites=true;
				moveto(m_pathxdest*8, m_pathydest*8, m_movespeed_num, m_movespeed_denom);
			}
			else
				nextstepinpath();
		}
	}

	object::compute(lowtime);

	if(m_turning)
	{
		if(!--m_turncountdown) // next step due?
		{
			setangle(m_angledeg+m_turnspeednumdeg);
			m_turntoangle-=abs(m_turnspeednum);
			m_turncountdown=m_turnspeeddenom;
		}

#ifdef DEBUG_TURNING
		uprintf("from %d/%d to %d\n", m_angle,m_angledeg, m_turntoangle);
#endif

		// done yet?
		if(m_turntoangle<=0) m_turning=false;
	}
#ifdef PROFILE_SPRITE
	TOOL_PROFILER_END;
#endif
}

int astarqueue::m_delqueue=-1;
node* astarqueue::m_delqueues[A_MAXDELQUEUES] MEM_IN_EWRAM;

//#define DEBUG_DELETEQUEUES

void astarqueue::deleteme(node* queue)
{
	ASSERT(m_delqueue<A_MAXDELQUEUES-1, "max number of delete queues exceeded");
	m_delqueues[++m_delqueue]=queue;
#ifdef DEBUG_DELETEQUEUES
	uprintf("delqueue %d\n", m_delqueue);
#endif
}

bool astarqueue::deletethem(int maxtime)
{
	if(m_delqueue<0)
		return false;
	else
	{
#ifdef DEBUG_DELETEQUEUES
		int n=0;
#endif
		// Timer 2 is running (started by stage::compute())
		int time=hw::getproftimer();
		while(m_delqueue>-1 && hw::getproftimer()-time < maxtime)
		{
			if(m_delqueues[m_delqueue]==0) {
				m_delqueue--; continue;
			}
#ifdef DEBUG_DELETEQUEUES
			n++;
#endif
			node* d=m_delqueues[m_delqueue]->qparent;
			delete m_delqueues[m_delqueue];
			m_delqueues[m_delqueue]=d;
		}
#ifdef DEBUG_DELETEQUEUES
		uprintf("%d nodes deld\n", n);
#endif
	}
	return true;
}

void sprite::stop()
{
	if(m_resumepathing)
	{
		m_resumepathing=false;
		delete m_open;
		delete m_closed;
	}
	if(m_ispathing)
	{
		m_ispathing=false;
		if(m_path)
		{
			delete m_path;
			m_path=0;
		}
	}
	m_ismoving=false;
}

void sprite::enablerotation(bool rot)
{
	if(rot==m_rotatable) return;
	m_rotatable=rot;
	if(rot)	   // turn on rotation
	{
		m_rotset=getrotset();
		ham_SetObjDblSize(m_objno, 1);
		ham_SetObjRotEnable(m_objno, 1);
		ham_SetObjRotSetSelect(m_objno, m_rotset);
		setangle(0);
	}
	else
	{		 // turn off rotation
		m_turning=false;
		freerotset(m_rotset);
		ham_SetObjDblSize(m_objno, 0);
		ham_SetObjRotEnable(m_objno, 0);
	}
}

void sprite::setangle(int angle)
{
	ASSERT(m_rotatable, "not a rot sprite!");
	if(angle<0)	angle+=360;
	if(angle>359) angle-=360;
	ASSERT(angle>=0 && angle<360, "angle out of range");

	m_angle=BIOSDEGREES(angle);
	m_angledeg=angle;
	ham_RotObjDefineSet(m_rotset, angle, 0x100, 0x100);
}

void sprite::turnto(int angle, int speednum, int speeddenom)
{
	ASSERT(m_rotatable, "not a rot sprite!");

	u16 a=BIOSDEGREES(angle);

	if(a==m_angle || speednum==0)  return;

	if(speednum<0)
	{
		if(a>m_angle)
			m_turntoangle=m_angle+0x10000-a;
		else
			m_turntoangle=m_angle-a;
	}
	else
	{
		if(a<m_angle)
			m_turntoangle=a+0x10000-m_angle;
		else
			m_turntoangle=a-m_angle;
	}
	m_turning=true;
	m_turnspeednum=BIOSDEGREES(speednum);
	m_turnspeednumdeg=speednum;
	m_turnspeeddenom=m_turncountdown=speeddenom;
}

void sprite::setpriority(int prio)
{
	if(prio==P_ONTOP) ham_SetObjPrio(m_objno, 0);
	else ham_SetObjPrio(m_objno, 3-(m_parent->m_first_bg+prio));
}

void sprite::setobjbefore(sprite* other)
{
	ham_SetObjBefore(m_objno, other->m_objno);
}

#define SHOVESPEED 2

void sprite::push(sprite* pusher)
{
	if(m_shovable)
	{
		int dx=getboundxpos()-pusher->getboundxpos();
		int dy=getboundypos()-pusher->getboundypos();
		if(abs(dx)>abs(dy))
		{
			if(dx>0) setxpos(getxpos()+SHOVESPEED);
			else setxpos(getxpos()-SHOVESPEED);
		}
		else
		{
			if(dy>0) setypos(getypos()+SHOVESPEED);
			else setypos(getypos()-SHOVESPEED);
		}
	}
}

// -------- rotset handling

bool sprite::m_rotsets[MAX_ROTSETS] MEM_IN_EWRAM;

int sprite::getrotset()
{
	for(int i=0; i<MAX_ROTSETS; i++)
		if(!m_rotsets[i])
		{
			m_rotsets[i]=true;
			return i;
		}

	ASSERT(0, "out of rotsets");
	return -1;
}

void sprite::freerotset(int rotset)
{
	m_rotsets[rotset]=false;
}

void sprite::setpalette(const palettedata* palette)
{
	ham_LoadObjPal((void*)(palette->palette), palette->entries);
}

void sprite::setpalette16(int palno, const palettedata* palette)
{
	ham_LoadObjPal16((void*)(palette->palette), palno);
}

